from joonmyung.meta_data.label import imnet_label, cifar_label
from timm.data import create_dataset, create_loader
from torchvision import transforms
from joonmyung.utils import getDir
import torch
import copy
import glob
import PIL
import os


# CIFAR Setting
# pip install cifar2png
# cifar2png cifar100 ./cifar100
# cifar2png cifar10  ./cifar10

class JDataset():
    # transforms.Resize(int((256 / 224) * input_size), interpolation=InterpolationMode.BICUBIC),
    settings = {"imagenet" : {
                    "num_classes"   : 1000,
                    "data_types"    : ["val", "train"],
                    "label_name"    : imnet_label,
                    "distributions" : {"mean": [0.485, 0.456, 0.406], "std": [0.229, 0.224, 0.225]},
                    "size": (224, 224)
                    },
                "cifar100" : {
                    "num_classes" : 100,
                    "data_types": ["test", "train"],
                    "label_name" : cifar_label,
                    "distributions": {"mean": [0.4914, 0.4822, 0.4465], "std": [0.2023, 0.1994, 0.2010]},
                    "size" : (32, 32)
                    }
    }

    def __init__(self, data_path="/hub_data1/joonmyung/data", dataset="imagenet", train=False, transform_type = 0,
                 distribution = None, size = None, device="cuda"):
        self.dataset = dataset.lower()
        setting = self.settings[self.dataset]
        self.label_name     = setting["label_name"]
        self.data_type      = setting["data_types"][train]
        self.transform_type = transform_type
        self.distribution   = distribution if distribution else setting["distributions"]
        size                = size if size else setting["size"]

        self.transform = [
                transforms.Compose([transforms.Resize(size, interpolation=3), transforms.ToTensor(), transforms.Normalize(self.distribution["mean"], self.distribution["std"])]),
                transforms.Compose([transforms.Resize(size, interpolation=3), transforms.ToTensor()]),
                transforms.Compose([transforms.ToTensor()])
        ]

        self.device = device
        self.data_path = data_path
        self.label_paths = sorted(getDir(os.path.join(self.data_path, self.data_type)))
        self.img_paths   = [sorted(glob.glob(os.path.join(self.data_path, self.data_type, label_path, "*"))) for label_path in self.label_paths]
        # self.img_paths   = [sorted(glob.glob(os.path.join(self.data_path, self.data_type, "*", "*")))]


    def __getitem__(self, idx):
        if len(idx) == 2:
            label_num, img_num = idx
            transform_type = self.transform_type
        elif len(idx) == 3:
            label_num, img_num, transform_type = idx

        img_path = self.img_paths[label_num][img_num]
        img = PIL.Image.open(img_path)
        data = self.transform[transform_type](img)

        return data.unsqueeze(0).to(self.device), torch.tensor(label_num).to(self.device), self.label_name[int(label_num)]

    def getItems(self, indexs):
        ds, ls, lns = [], [], []
        for index in indexs:
            d, l, ln = self.__getitem__(index)
            ds.append(d)
            ls.append(l)
            lns.append(ln)
        return torch.cat(ds, dim=0), torch.stack(ls, dim=0), lns

    def getItemPath(self, img_path):
        img = PIL.Image.open(img_path)
        data = self.transform(img)
        return data.unsqueeze(0).to(self.device)

    def __len__(self):
        return

    def getAllItems(self, batch_size=32):
        dataset = create_dataset(
            root=self.data_path, name="IMNET"
            , split='validation', is_training=False
            , load_bytes=False, class_map='')

        loader = create_loader(
            dataset,
            input_size=(3, 224, 224),
            batch_size=batch_size,
            use_prefetcher=True,
            interpolation='bicubic',
            mean = self.distribution["mean"],
            std  = self.distribution["std"],
            num_workers=8,
            crop_pct=0.9,
            pin_memory=False,
            tf_preprocessing=False)
        return loader

    def validation(self, data):
        return data.lower()

    def unNormalize(self, image):
        result = torch.zeros_like(image)
        for c, (m, s) in enumerate(zip(self.distribution["mean"], self.distribution["std"])):
            result[:, c] = image[:, c].mul(s).add(m)
        return result

    def normalize(self, image):
        result = copy.deepcopy(image)
        for c, (m, s) in enumerate(zip(self.distribution["mean"], self.distribution["std"])):
            result[:, c].sub_(m).div_(s)
        return result

if __name__ == "__main__":
    root_path = "/hub_data1/joonmyung/data"
    dataset = "imagenet"
    dataset = JDataset(root_path, dataset, train=True, )
    d, l, l_n  = dataset[0, 1]
    samples = dataset.getitems([[0,1], [0,2], [0,3]])
    print(1)



import json
import os
import pickle
codecs = ['CP949', 'utf8', 'euc-kr',
          'ascii', 'big5', 'big5hkscs', 'cp037', 'cp273', 'cp424', 'cp437', 'cp500', 'cp720', 'cp737', 'cp775', 'cp850', 'cp852', 'cp855',
          'cp856', 'cp857', 'cp858', 'cp860', 'cp861', 'cp862', 'cp863', 'cp864', 'cp865', 'cp866', 'cp869', 'cp874', 'cp875', 'cp932', 'cp949',
          'cp950', 'cp1006', 'cp1026', 'cp1125', 'cp1140', 'cp1250', 'cp1251', 'cp1252', 'cp1253', 'cp1254', 'cp1255', 'cp1256', 'cp1257', 'cp1258',
          'euc_jp', 'euc_jis_2004', 'euc_jisx0213', 'euc_kr', 'gb2312', 'gbk', 'gb18030', 'hz', 'iso2022_jp', 'iso2022_jp_1', 'iso2022_jp_2',
          'iso2022_jp_2004', 'iso2022_jp_3', 'iso2022_jp_ext', 'iso2022_kr', 'latin_1', 'iso8859_2', 'iso8859_3', 'iso8859_4', 'iso8859_5', 'iso8859_6',
          'iso8859_7', 'iso8859_8', 'iso8859_9', 'iso8859_10', 'iso8859_11', 'iso8859_13', 'iso8859_14', 'iso8859_15', 'iso8859_16', 'johab', 'koi8_r', 'koi8_t',
          'koi8_u', 'kz1048', 'mac_cyrillic', 'mac_greek', 'mac_iceland', 'mac_latin2', 'mac_roman', 'mac_turkish', 'ptcp154', 'shift_jis', 'shift_jis_2004',
          'shift_jisx0213', 'utf_32', 'utf_32_be', 'utf_32_le', 'utf_16', 'utf_16_be', 'utf_16_le', 'utf_7', 'utf_8', 'utf_8_sig']


def read(file_path, d1, condition_record, c_sites, p=False):
    filetype = file_path.split(".")[-1].lower()
    if os.path.isfile(file_path):
        with open(file_path, "rb") as f:
            if filetype == "json":
                return json.load(f)
            elif filetype =="csv":
                for codec in codecs:
                    try:
                        csv_df = pd.read_csv(file_path
                                             , header=0
                                             , encoding=codec
                                             # , sep=',|\s+|;|:'
                                             , sep=','
                                             # , iterator=True, chunksize=10000 # b. -1~0% (73~74초 → 74초), 파일 사이즈가 큰 경우 유리
                                             # , dtype=None
                                             # , usecols = cols
                                             , )
                    except:
                        continue
                    csv_df = condition(columnControl(csv_df, d1), condition_record, p=p)

                    return csv_df

                raise Exception("Encoding Error for {}".format(file_path))

            elif filetype == "mat":
                from scipy import io
                return io.loadmat(file_path)
            elif filetype == "xml":
                pass
            elif filetype == "pickle":
                return pickle.load(f)
    else:
        return False



import numpy as np
import GPUtil
import random
import torch
import os


def setGPU(gpuNum):
    os.environ["CUDA_VISIBLE_DEVICES"] = gpuNum
    print('Available devices ', torch.cuda.device_count())
    print('Current cuda device ', torch.cuda.current_device())


def selectGPU(gpuNum, p=True):
    os.environ["CUDA_VISIBLE_DEVICES"] = str(gpuNum)
    if p:
        print('Available devices ', torch.cuda.device_count())
        print('Current cuda device ', torch.cuda.current_device())


def fixSeed(seed, fast=False, p = True):
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)

    if fast:
        torch.backends.cudnn.deterministic = False
        torch.backends.cudnn.benchmark = True
    else:
        # use_deterministic_algorithms : deterministic보다 더 넓은 범위 커버, non-deterministic한 알고리즘이 존재하는 경우 Error 발생
        # pytorch 1.8.0부터 사용
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False
        if torch.__version__ > "1.8.0":
            torch.use_deterministic_algorithms(True)

    if p: print("torch : {} \tcuda : {} \tnumpy : {}".format(torch.initial_seed(), torch.cuda.initial_seed(), np.random.get_state()[1][0]))


def on_terminate(proc):
    print("process {} terminated".format(proc))


def gpuUtil():
    GPUtil.showUtilization()


import pickle

from joonmyung.utils import is_dist_avail_and_initialized
from collections import defaultdict, deque
from joonmyung.draw import data2PIL
from joonmyung.utils import to_np
import torch.distributed as dist
import datetime
import wandb
import torch
import time
import os

class AverageMeter:
    ''' Computes and stores the average and current value. '''
    def __init__(self) -> None:
        self.reset()

    def reset(self) -> None:
        self.val = 0.0
        self.avg = 0.0
        self.sum = 0.0
        self.count = 0

    def update(self, val: float, n: int = 1) -> None:
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count
    def __str__(self):
        return "\
        end = time.time() \n\
        batch_time = AverageMeter() \n\
        batch_time.update(time.time() - end) \n\
        end = time.time() \n\
        avg_score = AverageMeter()\n\
        accuracy = 0.1\n\
        avg_score.update(accuracy)\n\
        losses = AverageMeter()\n\
        loss = 0\n\
        batch_size = 128\n\
        losses.update(loss.data.item(), batch_size)\n\
        print(f'time {batch_time.val:.3f} ({batch_time.avg:.3f})\t'\n\
              f'loss {losses.val:.4f} ({losses.avg:.4f})\t' \n\
              f'acc {avg_score.val:.4f} ({avg_score.avg:.4f})')"




class Logger():
    loggers = {}
    def __init__(self, use_wandb=True, wandb_entity=None, wandb_project=None, wandb_name=None
                 , wandb_watch=False, main_process=True, wandb_id=None, wandb_dir='./'
                 , args=None, model=False):
        self.use_wandb = use_wandb
        self.main_process = main_process

        if self.use_wandb and self.main_process:
            wandb.init(entity=wandb_entity, project=wandb_project, name=wandb_name
                       , save_code=True, resume="allow", id = wandb_id, dir=wandb_dir
                       , config=args, settings=wandb.Settings(code_dir="."))

            if args:
                args.wandb_id = wandb.config.id = wandb.run.id
                torch.save(args, os.path.join(wandb.run.dir, "args.pt"))
            if wandb_watch and model: wandb.watch(model, log='all')


    def getLog(self, k, return_type =None):
        if return_type == "avg":
            return self.loggers[k].avg
        elif return_type == "val":
            return self.loggers[k].val
        else:
            return self.loggers[k]

    def delLog(self, columns: list):
        for column in columns:
            self.loggers.pop(column)

    def resetLog(self):
        self.loggers = {k:AverageMeter() if type(v) == AverageMeter else v for k, v in self.loggers.items()}

    def addLog(self, datas:dict, epoch=None, bs = 1):
        if self.main_process and self.use_wandb:
            for k, v in datas.items():
                data_type = v[0]
                if data_type == 0:  # Values
                    self.loggers[k] = v[1]
                elif data_type == 1: # AverageMeter
                    if k not in self.loggers.keys():
                        self.loggers[k] = AverageMeter()
                    self.loggers[k].update(v[1], bs)
                elif data_type == 2: # Table
                    columns = list(v[1].keys())
                    data_num = len(list(v[1].values())[0])
                    self.loggers[k] = wandb.Table(columns=["epoch"] + columns)
                    for idx in range(data_num):
                        results = []
                        for c in columns:
                            if type(v[1][c]) == torch.Tensor:
                                result = wandb.Image(to_np(data2PIL(v[1][c][idx]))) if len(v[1][c].shape) == 4 else to_np(v[1][c])[idx]
                            elif type(v[1][c]) == list:
                                result = v[1][c][idx]
                            else:
                                raise ValueError
                            results.append(result)
                        self.loggers[k].add_data(str(epoch), *results)
        return True

    def getPath(self):
        return wandb.run.dir

    def logWandb(self):
        if self.main_process and self.use_wandb:
            wandb.log({k: v.avg if type(v) == AverageMeter else v for k, v in self.loggers.items()})
            self.resetLog()

    def finish(self):
        wandb.finish()

    def save(self, model, args, name):
        if self.main_process:
            path = os.path.join(wandb.run.dir, f"{name}.pth")
            torch.save({"model" : model, "args"  : args}, path)
            wandb.save(path, wandb.run.dir)


class SmoothedValue(object):
    """Track a series of values and provide access to smoothed values over a
    window or the global series average.
    """

    def __init__(self, window_size=20, fmt=None):
        if fmt is None:
            fmt = "{median:.4f} ({global_avg:.4f})"
        self.deque = deque(maxlen=window_size)
        self.total = 0.0
        self.count = 0
        self.fmt = fmt

    def update(self, value, n=1):
        self.deque.append(value)
        self.count += n
        self.total += value * n

    def synchronize_between_processes(self):
        """
        Warning: does not synchronize the deque!
        """
        if not is_dist_avail_and_initialized():
            return
        t = torch.tensor([self.count, self.total], dtype=torch.float64, device='cuda')
        dist.barrier()
        dist.all_reduce(t)
        t = t.tolist()
        self.count = int(t[0])
        self.total = t[1]

    @property
    def median(self):
        d = torch.tensor(list(self.deque))
        return d.median().item()

    @property
    def avg(self):
        d = torch.tensor(list(self.deque), dtype=torch.float32)
        return d.mean().item()

    @property
    def global_avg(self):
        return self.total / self.count

    @property
    def max(self):
        return max(self.deque)

    @property
    def value(self):
        return self.deque[-1]

    def __str__(self):
        return self.fmt.format(
            median=self.median,
            avg=self.avg,
            global_avg=self.global_avg,
            max=self.max,
            value=self.value)


class MetricLogger(object):
    def __init__(self, delimiter="\t"):
        self.meters = defaultdict(SmoothedValue)
        self.delimiter = delimiter

    def update(self, **kwargs):
        for k, v in kwargs.items():
            if isinstance(v, torch.Tensor):
                v = v.item()
            assert isinstance(v, (float, int))
            self.meters[k].update(v)

    def __getattr__(self, attr):
        if attr in self.meters:
            return self.meters[attr]
        if attr in self.__dict__:
            return self.__dict__[attr]
        raise AttributeError("'{}' object has no attribute '{}'".format(
            type(self).__name__, attr))

    def __str__(self):
        loss_str = []
        for name, meter in self.meters.items():
            loss_str.append(
                "{}: {}".format(name, str(meter))
            )
        return self.delimiter.join(loss_str)

    def synchronize_between_processes(self):
        for meter in self.meters.values():
            meter.synchronize_between_processes()

    def add_meter(self, name, meter):
        self.meters[name] = meter

    def log_every(self, iterable, print_freq, header=None):
        i = 0
        if not header:
            header = ''
        start_time = time.time()
        end = time.time()
        iter_time = SmoothedValue(fmt='{avg:.4f}')
        data_time = SmoothedValue(fmt='{avg:.4f}')
        space_fmt = ':' + str(len(str(len(iterable)))) + 'd'
        log_msg = [
            header,
            '[{0' + space_fmt + '}/{1}]',
            'eta: {eta}',
            '{meters}',
            'time: {time}',
            'data: {data}'
        ]
        if torch.cuda.is_available():
            log_msg.append('max mem: {memory:.0f}')
        log_msg = self.delimiter.join(log_msg)
        MB = 1024.0 * 1024.0
        for obj in iterable:
            data_time.update(time.time() - end)
            yield obj
            iter_time.update(time.time() - end)
            if i % print_freq == 0 or i == len(iterable) - 1:
                eta_seconds = iter_time.global_avg * (len(iterable) - i)
                eta_string = str(datetime.timedelta(seconds=int(eta_seconds)))
                if torch.cuda.is_available():
                    print(log_msg.format(
                        i, len(iterable), eta=eta_string,
                        meters=str(self),
                        time=str(iter_time), data=str(data_time),
                        memory=torch.cuda.max_memory_allocated() / MB))
                else:
                    print(log_msg.format(
                        i, len(iterable), eta=eta_string,
                        meters=str(self),
                        time=str(iter_time), data=str(data_time)))
            i += 1
            end = time.time()
        total_time = time.time() - start_time
        total_time_str = str(datetime.timedelta(seconds=int(total_time)))
        print('{} Total time: {} ({:.4f} s / it)'.format(
            header, total_time_str, total_time / len(iterable)))




from pathlib import Path
def list2name(value):
    return str(value).replace(" ", "").replace("],[", "_").replace("[", "").replace(",", "").replace("]", "")
def list2str(value):
    return str(value).replace(" ", "")


def resultManager(file_name, folder_path = "./result",
                  new_result = None, checkColumns = None, duplicate = False, duplicate_check=False):
    Path(folder_path).mkdir(parents=True, exist_ok=True)
    file_path = os.path.join(folder_path, file_name)
    try:
        with open(file_path, 'rb') as picklefile:
            results = pickle.load(picklefile)
    except FileNotFoundError:
        results = []

    if new_result is not None:  # update log file
        if checkColumns is not None and (not duplicate or duplicate_check):  # check exist
            for result in results:
                # 하나도 안 같은경우
                if not sum([list2str(result[checkColumn]) != list2str(new_result[checkColumn]) if checkColumn in result.keys() else True for checkColumn in checkColumns]):
                    return False
            if duplicate_check:
                return True
        results.append(new_result)
    else:
        return results

    with open(file_path, 'wb') as picklefile: # pass
        pickle.dump(results, picklefile)
    print(f"saved result path : {file_path}")
    return True





# if __name__ == "__main__":
#     from playground.analysis.lib_import import *
#     dataset_name, server, device = "imagenet", "148", "cuda"
#     data_path, _ = data2path(server, dataset_name)
#     data_num = [[5, 1],
#                 [5, 2],
#                 [5, 3],
#                 [5, 4]]
#     dataset = JDataset(data_path, dataset_name, device=device)
#     samples, targets, imgs, label_names = dataset.getItems(data_num)
#     model = torch.hub.load('facebookresearch/deit:main', "deit_tiny_patch16_224", pretrained=True)
#
#     logger = Logger(use_wandb=True, wandb_entity="joonmyung", wandb_project="test", wandb_name="AAPP",
#                     wandb_watch=False, wandb_dir="./")
#
#     logger.addLog({ "sample A": [0, 4],
#                     "sample B": [0, 3],
#                     "sample C": [0, 2],
#                     "table  B": [2, {"image" :     samples, "prediction": targets}]})
#
#
#     logger.save(model.state_dict(), "checkpoint_best.pt")
#     logger.logWandb()
#     logger.finish()


import os
os.environ["CUDA_VISIBLE_DEVICES"] = "1"
from joonmyung.analysis.dataset import JDataset
from joonmyung.analysis.model import JModel
from joonmyung.draw import saliency, overlay, drawImgPlot, drawHeatmap, unNormalize
from joonmyung.meta_data import data2path
from joonmyung.data import getTransform
from joonmyung.metric import targetPred
from joonmyung.log import AverageMeter
from joonmyung.utils import to_leaf, to_np
from tqdm import tqdm
from contextlib import suppress
import matplotlib.pyplot as plt
import torch.nn.functional as F
import numpy as np
import torch
import PIL
import cv2


def anaModel(transformer_class):
    class VisionTransformer(transformer_class):
        def forward_features(self, x):
            x = self.patch_embed(x)
            cls_token = self.cls_token.expand(x.shape[0], -1, -1)  # stole cls_tokens impl from Phil Wang, thanks
            if self.dist_token is None:
                x = torch.cat((cls_token, x), dim=1)
            else:
                x = torch.cat((cls_token, self.dist_token.expand(x.shape[0], -1, -1), x), dim=1)


            if self.analysis[0] == 1:   # PATCH
                x = x # (8, 197, 192)
            elif self.analysis[0] == 2: # POS
                x = self.pos_embed # (1, 197, 192)
            elif self.analysis[0] == 3:  # PATCH (RANDOM I) + POS
                x = torch.rand_like(self.pos_embed, device=x.device) + self.pos_embed
            elif self.analysis[0] == 4:  # PATCH (RANDOM II) + POS
                x = torch.rand_like(self.cls_token, device=x.device).repeat(1, x.shape[1], 1) + self.pos_embed
            else: # PATCH + POS
                x = x + self.pos_embed
            x = self.pos_drop(x)

            x = self.blocks(x)
            x = self.norm(x)
            if self.dist_token is None:
                return self.pre_logits(x[:, 0])
            else:
                return x[:, 0], x[:, 1]

    return VisionTransformer

class Analysis:
    def __init__(self, model, analysis = [0], activate = [True, False, False], detach=True, key_name=None, num_classes = 1000
                 , cls_start=0, cls_end=1, patch_start=1, patch_end=None
                 , ks = 5
                 , amp_autocast=suppress, device="cuda"):
        # Section A. Model
        self.num_classes = num_classes
        self.key_name = key_name

        model_ = anaModel(model.__class__)
        model.__class__ = model_
        model.analysis = analysis
        self.model = model
        self.ks = ks
        self.detach = detach

        # Section B. Attention
        self.kwargs_roll = {"cls_start" : cls_start, "cls_end" : cls_end,
                            "patch_start" : patch_start, "patch_end" : patch_end}


        # Section C. Setting
        hooks = [{"name_i": 'attn_drop', "name_o": 'decoder', "fn_f": self.attn_forward, "fn_b": self.attn_backward},
                 {"name_i": 'qkv', "name_o": 'decoder', "fn_f": self.qkv_forward, "fn_b": self.qkv_backward},
                 {"name_i": 'head', "name_o": 'decoder', "fn_f": self.head_forward, "fn_b": self.head_backward}]
        hooks = [h for h, a in zip(hooks, activate) if a]


        self.amp_autocast = amp_autocast
        self.device       = device

        for name, module in self.model.named_modules():
            for hook in hooks:
                if hook["name_i"] in name and hook["name_o"] not in name:
                    module.register_forward_hook(hook["fn_f"])
                    module.register_backward_hook(hook["fn_b"])

    def attn_forward(self, module, input, output):
        # input  : 1 * (8, 3, 197, 197)
        # output : (8, 3, 197, 197)
        self.info["attn"]["f"].append(output.detach() if self.detach else output)

    def attn_backward(self, module, grad_input, grad_output):
        # input  : 1 * (8, 3, 197, 192)
        # output : (8, 3, 197, 576)
        self.info["attn"]["b"].insert(0, grad_input[0].detach() if self.detach else grad_input[0])

    def qkv_forward(self, module, input, output):
        # input  : 1 * (8, 197, 192)
        # output : (8, 197, 576)
        self.info["qkv"]["f"].append(output.detach())

    def qkv_backward(self, module, grad_input, grad_output):
        # self.info["qkv"]["b"].append(grad_input[0].detach())
       pass

    def head_forward(self, module, input, output):
        # input : 1 * (8(B), 192(D)), output : (8(B), 1000(C))
        # TP = targetPred(output.detach(), self.targets.detach(), topk=self.ks)
        TP = targetPred(to_leaf(output), to_leaf(self.targets), topk=self.ks)
        self.info["head"]["TP"] = torch.cat([self.info["head"]["TP"], TP], dim=0) if "TP" in self.info["head"].keys() else TP

    def head_backward(self, module, grad_input, grad_output):
        pass

    def resetInfo(self):
        self.info = {"attn": {"f": [], "b": []}, "qkv": {"f": [], "b": []},
                     "head": {"acc1" : AverageMeter(),
                              "acc5" : AverageMeter(),
                              "pred" : None

                              }}

    def __call__(self, samples, index=None, **kwargs):
        self.resetInfo()
        self.model.zero_grad()
        self.model.eval()

        if type(samples) == torch.Tensor:
            outputs = self.model(samples, **kwargs)
            return outputs
        else:
            for sample, self.targets in tqdm(samples):
                _ = self.model(sample)
            return False

    def anaSaliency(self, attn=True, grad=False, output=None, index=None,
                    head_fusion="mean", discard_ratios=[0.], data_from="cls",
                    ls_attentive=[], ls_rollout=[],
                    reshape=False, device="cuda"):

        if attn:
            attn = self.info["attn"]["f"]
        if grad:
            self.info["attn"]["b"] = []
            self.model.zero_grad()
            if index == None: index = output.max(dim=1)[1]
            index = torch.eye(self.num_classes, device=self.device)[index]
            loss = (output * index).sum()
            loss.backward(retain_graph=True)
            grad = self.info["attn"]["b"]

        return saliency(attn, grad
                        , head_fusion=head_fusion, discard_ratios=discard_ratios, data_from=data_from
                        , ls_rollout=ls_rollout, ls_attentive=ls_attentive, reshape=reshape, device=device)


if __name__ == '__main__':
    # Section A. Data
    dataset_name, device, amp_autocast, debug = "imagenet", 'cuda', torch.cuda.amp.autocast, True
    data_path, num_classes, _, _ = data2path(dataset_name)
    view, activate = [False, True, False, False, False], [True, False, False]
        # VIEW     : IMG, SALIENCY:ATTN, SALIENCY:OPENCV, SALIENCY:GRAD, ATTN. MOVEMENT
        # ACTIVATE : ATTN, QKV, HEAD
    analysis = [0] # [0] : INPUT TYPE, [0 : SAMPLE + POS, 1 : SAMPLE, 2 : POS]

    dataset = JDataset(data_path, dataset_name, device=device)
    data_idxs = [[c, i] for i in range(1000) for c in range(50)]
    # data_idxs = [[21, 0], [22, 0], [2, 0], [0, 0], [0, 1], [1, 1], [2, 1], [3, 1]]

    # Section B. Model
    model_number, model_name = 0, "deit_tiny_patch16_224" # deit, vit | tiny, small, base
    # model_number, model_name = 1, "deit_tiny_patch16_224"

    # Section C. Setting
    ls_rollout, ls_attentive, col = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11], [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11], 12

    modelMaker = JModel(num_classes, device=device)
    model = modelMaker.getModel(model_number, model_name)
    model = Analysis(model, analysis = analysis, activate = activate, device=device)
    for idx, data_idx in enumerate(data_idxs):
        print(f"------------------------- [{data_idx[0]}]/[{data_idx[1]}] -------------------------")

        sample, target, label_name = dataset[data_idx[0], data_idx[1]]
        # sample, _, img, _ = dataset.getItemPath('/hub_data1/joonmyung/data/imagenet/train/n01440764/n01440764_39.JPEG')
        output = model(sample)
        if view[0]:
            drawImgPlot(unNormalize(sample, "imagenet"))

        if view[1]: # SALIENCY W/ MODEL
            # ls_rollout, ls_attentive, col = [], [0,2,4,6,8,10], 6
            # discard_ratios, v_ratio, head_fusion, data_from = [0.0, 0.4, 0.8], 0.1, "mean", "cls"
            discard_ratios, v_ratio, head_fusion, data_from = [0.0], 0.0, "mean", "patch"  # Attention, Gradient
            rollout, attentive = model.anaSaliency(True, False, output, discard_ratios=discard_ratios,
                                                   ls_attentive = ls_attentive, ls_rollout=ls_rollout,
                                                   head_fusion  = head_fusion, index=target, data_from=data_from,
                                                   reshape      = True) # (12(L), 8(B), 14(H), 14(W))
            datas_attn    = overlay(sample, attentive, dataset_name)
            drawImgPlot(datas_attn, col=col)

            discard_ratios, v_ratio, head_fusion, data_from = [0.0], 0.0, "mean", "cls"  # Attention, Gradient
            rollout, attentive = model.anaSaliency(True, False, output, discard_ratios=discard_ratios,
                                                   ls_attentive = ls_attentive, ls_rollout=ls_rollout,
                                                   head_fusion  = head_fusion, index=target, data_from=data_from,
                                                   reshape      = True) # (12(L), 8(B), 14(H), 14(W))
            datas_attn = overlay(sample, attentive, dataset_name)
            drawImgPlot(datas_attn, col=col)


            # datas_rollout = overlay(sample, rollout,   dataset_name)
            # drawImgPlot(datas_rollout, col=col)

            # datas_attn = overlay(sample, attentive, dataset_name)
            # drawImgPlot(datas_attn, col=col)

            # a = attentive[5]
            # b = torch.stack([a.clamp(max=a.quantile(1 - v_ratio)) for v_ratio in [0.0, 0.05, 0.1, 0.15, 0.2, 0.25, 0.3, 0.35, 0.4, 0.45, 0.5, 0.55]])
            # datas_attn    = overlay(sample, b, dataset_name)
            # drawImgPlot(datas_attn, col=col)
            # print(1)

        if view[2]:  # SALIENCY W/ DATA
            img = np.array(dataset[data_idx[0], data_idx[1], 2][0])

            saliency = cv2.saliency.StaticSaliencySpectralResidual_create()
            (success, saliencyMap) = saliency.computeSaliency(img)
            saliencyMap = (saliencyMap * 255).astype("uint8")

            saliency = cv2.saliency.StaticSaliencyFineGrained_create()
            (success, saliencyFineMap) = saliency.computeSaliency(img)
            threshMap = cv2.threshold((saliencyFineMap * 255).astype("uint8"), 0, 255, cv2.THRESH_BINARY | cv2.THRESH_OTSU)[1]
            # plt.imshow(threshMap)
            # plt.show()

        if view[3]:  # SALIENCY FOR INPUT
            sample.requires_grad, model.detach, k = True, False, 3
            output = model(sample)
            attn = torch.stack(model.info["attn"]["f"], dim=1).mean(dim=[2,3])[0,-2]
            topK = attn[1:].topk(k, -1, True)[1]
            # a = torch.autograd.grad(attn.sum(), samples, retain_graph=True)[0].sum(dim=1)
            a = torch.autograd.grad(output[:,3], sample, retain_graph=True)[0].sum(dim=1)
            b = F.interpolate(a.unsqueeze(0), scale_factor=0.05, mode='nearest')[0]
            # drawHeatmap(b)
            print(1)
            # to_np(torch.stack([attn[:, :, 0], attn[:, :, 1:].sum(dim=-1)], -1)[0])

        if view[4]: # ATTENTION MOVEMENT (FROM / TO)
            attn = torch.stack(model.info["attn"]["f"]).mean(dim=2).transpose(0,1) # (8 (B), 12 (L), 197(T_Q), 197(T_K))
            cls2cls     = attn[:, :, :1, 0].mean(dim=2)              # (8(B), 12(L))
            patch2cls   = attn[:, :, :1, 1:].mean(dim=2).sum(dim=-1) # (8(B), 12(L))
            # PATCH가 받는 정도
            cls2patch   = attn[:, :, 1:, 0].mean(dim=2)
            patch2patch = attn[:, :, 1:, 1:].mean(dim=2).sum(dim=-1)
            # to_np(torch.stack([cls2cls.mean(dim=0), patch2cls.mean(dim=0), cls2patch.mean(dim=0), patch2patch.mean(dim=0)]))
            print(1)



from joonmyung.utils import time2str
from tqdm import tqdm
import subprocess
import time
import pynvml
import datetime

class GPU_Worker():
    def __init__(self, gpus:list, waitTimeInit = 30, waitTime = 60, count = 0,
                 checkType:int = 0, utilRatio:int = 50, need_gpu=1, max_run_num = 8, p = True):
        self.activate  = False

        self.gpus      = [int(gpu) for gpu in gpus]
        self.waitTimeInit = waitTimeInit
        self.waitTime = waitTime
        self.checkType = checkType
        self.need_gpu = int(need_gpu)
        self.max_run_num = max_run_num

        self.utilRatio = utilRatio
        self.p = p
        self.count = count

        self.availGPUs = []
        self.runGPUs   = {}

    def getFreeRatio(self, id):
        handle = pynvml.nvmlDeviceGetHandleByIndex(id)
        use = pynvml.nvmlDeviceGetUtilizationRates(handle)
        ratio = 0.5 * (float(use.gpu + float(use.memory)))
        # ratio = float(use.memory)
        # ratio = float(use.gpu)
        return ratio

    def setGPU(self):
        if self.activate: time.sleep(self.waitTimeInit)
        else: self.activate = True


        count = self.count
        pynvml.nvmlInit()
        while True:
            self.check_process()
            availGPUs = []
            count += 1
            for gpu in sorted(self.gpus - self.runGPUs.keys()):
                handle = pynvml.nvmlDeviceGetHandleByIndex(int(gpu))
                # 1. 아무것도 돌지 않는 경우
                if self.checkType == 0 and len(pynvml.nvmlDeviceGetComputeRunningProcesses(handle)) == 0:
                    availGPUs.append(gpu)

                # 2. n% 이하를 사용하고 있는 경우
                elif self.checkType == 1 and self.getFreeRatio(int(gpu)) < self.utilRatio:
                    availGPUs.append(gpu)

                if self.max_run_num <= len(availGPUs):
                    break

            if len(availGPUs) < self.need_gpu:
                if self.p:
                    print(f"{count}  |  CURRENT RUNNING GPUS : {list(self.runGPUs.keys())}")
                time.sleep(self.waitTime)
            else:
                break

        self.availGPUs = availGPUs
        if self.p: print("Activate GPUS : ", self.availGPUs)

    def getGPU(self):
        if len(self.availGPUs) < self.need_gpu: self.setGPU()
        gpus, self.availGPUs = self.availGPUs[:self.need_gpu], self.availGPUs[self.need_gpu:]

        return ','.join(map(str, gpus))


    def check_process(self):
        endGPUs = []
        for gpu, process in self.runGPUs.items():
            if process.poll() is not None:
                endGPUs.append(gpu)

        self.runGPUs = {gpu: p for gpu, p in sorted(self.runGPUs.items()) if gpu not in endGPUs}


    def register_process(self, gpus, p):
        for gpu in gpus.split(','):
            self.runGPUs[int(gpu)] = p

def Process_Worker(processes, gpuWorker, p = True):
    # TODO : 실험이 완전히 끝난 시간 체크할 필요가 존재함
    start = time.localtime()
    print("------ Start Running!! : {} ------".format(time2str(start)))

    for i, process in enumerate(tqdm(processes)):
        gpus = gpuWorker.getGPU()
        prefix = f"CUDA_VISIBLE_DEVICES={gpus} nohup sh -c \'"
        suffix = f"\' > {i+1}:gpu{gpus}.log 2>&1 "
        # print("------ {}:GPU{}  {} ------".format(i + 1, gpus, prefix + process + suffix))
        p = subprocess.Popen(prefix + process + suffix, shell=True)
        gpuWorker.register_process(gpus, p)

    end = time.localtime()
    print("------ End Running!!   : {} ------".format(time2str(end)))
    training_time = datetime.timedelta(seconds=time.mktime(end) - time.mktime(start))
    print(f"Time 1/all :  {training_time}/{training_time / len(processes)} ------")





if __name__ == '__main__':
    # Wokring Sample
    processes = [1,2,3,4,5]
    gpuWorker = GPU_Worker([0,1,2,3], 30, 120, checkType=1, utilRatio=50, need_gpu=4)
    Process_Worker(processes, gpuWorker)


import torch

def rangeBlock(block, vmin=0, vmax=5):
    loss = torch.arange(vmin, vmax, (vmax - vmin) / block, requires_grad=False).unsqueeze(dim=1)
    return loss

def columnRename(df, ns):
    for n in ns:
        if n[0] in df.columns:
            df.rename(columns = {n[0]: n[1]}, inplace = True)
#     columnRemove(df, ['c1', 'c2' ... ])


def columnRemove(df, ns):
    delList = []
    for n in ns:
        if n in df.columns:
            delList.append(n)
    df.drop(delList, axis=1, inplace=True)
#     columnRename(df, [['c1_p', 'c1_a'] , ['c2_p', 'c2_a']])


def normalization(t, type = 0):
    if type == 0:
        return t / t.max()
    elif type == 1:
        return t / t.min()

from torchvision import transforms
from torchvision.transforms import InterpolationMode

def getTransform(train = False, totensor = False, resize=True):

    if not resize:
        transform = lambda x: x
    else:
        transform = []

        transform.append(transforms.RandomResizedCrop(224, scale=(0.5, 1.0), interpolation=InterpolationMode.BICUBIC)) \
            if train else transform.append(transforms.Resize((224, 224), interpolation=3))

        if totensor:
            transform.append(transforms.ToTensor())
            transform.append(transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225]))
        transform = transforms.Compose(transform)

    return transform


import torch

from joonmyung.log import MetricLogger
from joonmyung.metric import accuracy


@torch.no_grad()
def evaluate(data_loader, model, device="cuda"):
    criterion = torch.nn.CrossEntropyLoss()

    metric_logger = MetricLogger(delimiter="  ")
    header = 'Test:'

    # switch to evaluation mode
    model.eval()

    for images, target in metric_logger.log_every(data_loader, 10, header):
        images = images.to(device, non_blocking=True)
        target = target.to(device, non_blocking=True)

        # compute output
        with torch.cuda.amp.autocast():
            output = model(images)
            loss = criterion(output, target)

        acc1, acc5 = accuracy(output, target, topk=(1, 5))

        batch_size = images.shape[0]
        metric_logger.update(loss=loss.item())
        metric_logger.meters['acc1'].update(acc1.item(), n=batch_size)
        metric_logger.meters['acc5'].update(acc5.item(), n=batch_size)

    print('* Acc@1 {top1.global_avg:.3f} Acc@5 {top5.global_avg:.3f} loss {losses.global_avg:.3f}'
          .format(top1=metric_logger.acc1, top5=metric_logger.acc5, losses=metric_logger.loss))

    return {k: meter.global_avg for k, meter in metric_logger.meters.items()}



imnet_label = {0: 'tench, Tinca tinca',
             1: 'goldfish, Carassius auratus',
             2: 'great white shark, white shark, man-eater, man-eating shark, Carcharodon carcharias',
             3: 'tiger shark, Galeocerdo cuvieri',
             4: 'hammerhead, hammerhead shark',
             5: 'electric ray, crampfish, numbfish, torpedo',
             6: 'stingray',
             7: 'cock',
             8: 'hen',
             9: 'ostrich, Struthio camelus',
             10: 'brambling, Fringilla montifringilla',
             11: 'goldfinch, Carduelis carduelis',
             12: 'house finch, linnet, Carpodacus mexicanus',
             13: 'junco, snowbird',
             14: 'indigo bunting, indigo finch, indigo bird, Passerina cyanea',
             15: 'robin, American robin, Turdus migratorius',
             16: 'bulbul',
             17: 'jay',
             18: 'magpie',
             19: 'chickadee',
             20: 'water ouzel, dipper',
             21: 'kite',
             22: 'bald eagle, American eagle, Haliaeetus leucocephalus',
             23: 'vulture',
             24: 'great grey owl, great gray owl, Strix nebulosa',
             25: 'European fire salamander, Salamandra salamandra',
             26: 'common newt, Triturus vulgaris',
             27: 'eft',
             28: 'spotted salamander, Ambystoma maculatum',
             29: 'axolotl, mud puppy, Ambystoma mexicanum',
             30: 'bullfrog, Rana catesbeiana',
             31: 'tree frog, tree-frog',
             32: 'tailed frog, bell toad, ribbed toad, tailed toad, Ascaphus trui',
             33: 'loggerhead, loggerhead turtle, Caretta caretta',
             34: 'leatherback turtle, leatherback, leathery turtle, Dermochelys coriacea',
             35: 'mud turtle',
             36: 'terrapin',
             37: 'box turtle, box tortoise',
             38: 'banded gecko',
             39: 'common iguana, iguana, Iguana iguana',
             40: 'American chameleon, anole, Anolis carolinensis',
             41: 'whiptail, whiptail lizard',
             42: 'agama',
             43: 'frilled lizard, Chlamydosaurus kingi',
             44: 'alligator lizard',
             45: 'Gila monster, Heloderma suspectum',
             46: 'green lizard, Lacerta viridis',
             47: 'African chameleon, Chamaeleo chamaeleon',
             48: 'Komodo dragon, Komodo lizard, dragon lizard, giant lizard, Varanus komodoensis',
             49: 'African crocodile, Nile crocodile, Crocodylus niloticus',
             50: 'American alligator, Alligator mississipiensis',
             51: 'triceratops',
             52: 'thunder snake, worm snake, Carphophis amoenus',
             53: 'ringneck snake, ring-necked snake, ring snake',
             54: 'hognose snake, puff adder, sand viper',
             55: 'green snake, grass snake',
             56: 'king snake, kingsnake',
             57: 'garter snake, grass snake',
             58: 'water snake',
             59: 'vine snake',
             60: 'night snake, Hypsiglena torquata',
             61: 'boa constrictor, Constrictor constrictor',
             62: 'rock python, rock snake, Python sebae',
             63: 'Indian cobra, Naja naja',
             64: 'green mamba',
             65: 'sea snake',
             66: 'horned viper, cerastes, sand viper, horned asp, Cerastes cornutus',
             67: 'diamondback, diamondback rattlesnake, Crotalus adamanteus',
             68: 'sidewinder, horned rattlesnake, Crotalus cerastes',
             69: 'trilobite',
             70: 'harvestman, daddy longlegs, Phalangium opilio',
             71: 'scorpion',
             72: 'black and gold garden spider, Argiope aurantia',
             73: 'barn spider, Araneus cavaticus',
             74: 'garden spider, Aranea diademata',
             75: 'black widow, Latrodectus mactans',
             76: 'tarantula',
             77: 'wolf spider, hunting spider',
             78: 'tick',
             79: 'centipede',
             80: 'black grouse',
             81: 'ptarmigan',
             82: 'ruffed grouse, partridge, Bonasa umbellus',
             83: 'prairie chicken, prairie grouse, prairie fowl',
             84: 'peacock',
             85: 'quail',
             86: 'partridge',
             87: 'African grey, African gray, Psittacus erithacus',
             88: 'macaw',
             89: 'sulphur-crested cockatoo, Kakatoe galerita, Cacatua galerita',
             90: 'lorikeet',
             91: 'coucal',
             92: 'bee eater',
             93: 'hornbill',
             94: 'hummingbird',
             95: 'jacamar',
             96: 'toucan',
             97: 'drake',
             98: 'red-breasted merganser, Mergus serrator',
             99: 'goose',
             100: 'black swan, Cygnus atratus',
             101: 'tusker',
             102: 'echidna, spiny anteater, anteater',
             103: 'platypus, duckbill, duckbilled platypus, duck-billed platypus, Ornithorhynchus anatinus',
             104: 'wallaby, brush kangaroo',
             105: 'koala, koala bear, kangaroo bear, native bear, Phascolarctos cinereus',
             106: 'wombat',
             107: 'jellyfish',
             108: 'sea anemone, anemone',
             109: 'brain coral',
             110: 'flatworm, platyhelminth',
             111: 'nematode, nematode worm, roundworm',
             112: 'conch',
             113: 'snail',
             114: 'slug',
             115: 'sea slug, nudibranch',
             116: 'chiton, coat-of-mail shell, sea cradle, polyplacophore',
             117: 'chambered nautilus, pearly nautilus, nautilus',
             118: 'Dungeness crab, Cancer magister',
             119: 'rock crab, Cancer irroratus',
             120: 'fiddler crab',
             121: 'king crab, Alaska crab, Alaskan king crab, Alaska king crab, Paralithodes camtschatica',
             122: 'American lobster, Northern lobster, Maine lobster, Homarus americanus',
             123: 'spiny lobster, langouste, rock lobster, crawfish, crayfish, sea crawfish',
             124: 'crayfish, crawfish, crawdad, crawdaddy',
             125: 'hermit crab',
             126: 'isopod',
             127: 'white stork, Ciconia ciconia',
             128: 'black stork, Ciconia nigra',
             129: 'spoonbill',
             130: 'flamingo',
             131: 'little blue heron, Egretta caerulea',
             132: 'American egret, great white heron, Egretta albus',
             133: 'bittern',
             134: 'crane',
             135: 'limpkin, Aramus pictus',
             136: 'European gallinule, Porphyrio porphyrio',
             137: 'American coot, marsh hen, mud hen, water hen, Fulica americana',
             138: 'bustard',
             139: 'ruddy turnstone, Arenaria interpres',
             140: 'red-backed sandpiper, dunlin, Erolia alpina',
             141: 'redshank, Tringa totanus',
             142: 'dowitcher',
             143: 'oystercatcher, oyster catcher',
             144: 'pelican',
             145: 'king penguin, Aptenodytes patagonica',
             146: 'albatross, mollymawk',
             147: 'grey whale, gray whale, devilfish, Eschrichtius gibbosus, Eschrichtius robustus',
             148: 'killer whale, killer, orca, grampus, sea wolf, Orcinus orca',
             149: 'dugong, Dugong dugon',
             150: 'sea lion',
             151: 'Chihuahua',
             152: 'Japanese spaniel',
             153: 'Maltese dog, Maltese terrier, Maltese',
             154: 'Pekinese, Pekingese, Peke',
             155: 'Shih-Tzu',
             156: 'Blenheim spaniel',
             157: 'papillon',
             158: 'toy terrier',
             159: 'Rhodesian ridgeback',
             160: 'Afghan hound, Afghan',
             161: 'basset, basset hound',
             162: 'beagle',
             163: 'bloodhound, sleuthhound',
             164: 'bluetick',
             165: 'black-and-tan coonhound',
             166: 'Walker hound, Walker foxhound',
             167: 'English foxhound',
             168: 'redbone',
             169: 'borzoi, Russian wolfhound',
             170: 'Irish wolfhound',
             171: 'Italian greyhound',
             172: 'whippet',
             173: 'Ibizan hound, Ibizan Podenco',
             174: 'Norwegian elkhound, elkhound',
             175: 'otterhound, otter hound',
             176: 'Saluki, gazelle hound',
             177: 'Scottish deerhound, deerhound',
             178: 'Weimaraner',
             179: 'Staffordshire bullterrier, Staffordshire bull terrier',
             180: 'American Staffordshire terrier, Staffordshire terrier, American pit bull terrier, pit bull terrier',
             181: 'Bedlington terrier',
             182: 'Border terrier',
             183: 'Kerry blue terrier',
             184: 'Irish terrier',
             185: 'Norfolk terrier',
             186: 'Norwich terrier',
             187: 'Yorkshire terrier',
             188: 'wire-haired fox terrier',
             189: 'Lakeland terrier',
             190: 'Sealyham terrier, Sealyham',
             191: 'Airedale, Airedale terrier',
             192: 'cairn, cairn terrier',
             193: 'Australian terrier',
             194: 'Dandie Dinmont, Dandie Dinmont terrier',
             195: 'Boston bull, Boston terrier',
             196: 'miniature schnauzer',
             197: 'giant schnauzer',
             198: 'standard schnauzer',
             199: 'Scotch terrier, Scottish terrier, Scottie',
             200: 'Tibetan terrier, chrysanthemum dog',
             201: 'silky terrier, Sydney silky',
             202: 'soft-coated wheaten terrier',
             203: 'West Highland white terrier',
             204: 'Lhasa, Lhasa apso',
             205: 'flat-coated retriever',
             206: 'curly-coated retriever',
             207: 'golden retriever',
             208: 'Labrador retriever',
             209: 'Chesapeake Bay retriever',
             210: 'German short-haired pointer',
             211: 'vizsla, Hungarian pointer',
             212: 'English setter',
             213: 'Irish setter, red setter',
             214: 'Gordon setter',
             215: 'Brittany spaniel',
             216: 'clumber, clumber spaniel',
             217: 'English springer, English springer spaniel',
             218: 'Welsh springer spaniel',
             219: 'cocker spaniel, English cocker spaniel, cocker',
             220: 'Sussex spaniel',
             221: 'Irish water spaniel',
             222: 'kuvasz',
             223: 'schipperke',
             224: 'groenendael',
             225: 'malinois',
             226: 'briard',
             227: 'kelpie',
             228: 'komondor',
             229: 'Old English sheepdog, bobtail',
             230: 'Shetland sheepdog, Shetland sheep dog, Shetland',
             231: 'collie',
             232: 'Border collie',
             233: 'Bouvier des Flandres, Bouviers des Flandres',
             234: 'Rottweiler',
             235: 'German shepherd, German shepherd dog, German police dog, alsatian',
             236: 'Doberman, Doberman pinscher',
             237: 'miniature pinscher',
             238: 'Greater Swiss Mountain dog',
             239: 'Bernese mountain dog',
             240: 'Appenzeller',
             241: 'EntleBucher',
             242: 'boxer',
             243: 'bull mastiff',
             244: 'Tibetan mastiff',
             245: 'French bulldog',
             246: 'Great Dane',
             247: 'Saint Bernard, St Bernard',
             248: 'Eskimo dog, husky',
             249: 'malamute, malemute, Alaskan malamute',
             250: 'Siberian husky',
             251: 'dalmatian, coach dog, carriage dog',
             252: 'affenpinscher, monkey pinscher, monkey dog',
             253: 'basenji',
             254: 'pug, pug-dog',
             255: 'Leonberg',
             256: 'Newfoundland, Newfoundland dog',
             257: 'Great Pyrenees',
             258: 'Samoyed, Samoyede',
             259: 'Pomeranian',
             260: 'chow, chow chow',
             261: 'keeshond',
             262: 'Brabancon griffon',
             263: 'Pembroke, Pembroke Welsh corgi',
             264: 'Cardigan, Cardigan Welsh corgi',
             265: 'toy poodle',
             266: 'miniature poodle',
             267: 'standard poodle',
             268: 'Mexican hairless',
             269: 'timber wolf, grey wolf, gray wolf, Canis lupus',
             270: 'white wolf, Arctic wolf, Canis lupus tundrarum',
             271: 'red wolf, maned wolf, Canis rufus, Canis niger',
             272: 'coyote, prairie wolf, brush wolf, Canis latrans',
             273: 'dingo, warrigal, warragal, Canis dingo',
             274: 'dhole, Cuon alpinus',
             275: 'African hunting dog, hyena dog, Cape hunting dog, Lycaon pictus',
             276: 'hyena, hyaena',
             277: 'red fox, Vulpes vulpes',
             278: 'kit fox, Vulpes macrotis',
             279: 'Arctic fox, white fox, Alopex lagopus',
             280: 'grey fox, gray fox, Urocyon cinereoargenteus',
             281: 'tabby, tabby cat',
             282: 'tiger cat',
             283: 'Persian cat',
             284: 'Siamese cat, Siamese',
             285: 'Egyptian cat',
             286: 'cougar, puma, catamount, mountain lion, painter, panther, Felis concolor',
             287: 'lynx, catamount',
             288: 'leopard, Panthera pardus',
             289: 'snow leopard, ounce, Panthera uncia',
             290: 'jaguar, panther, Panthera onca, Felis onca',
             291: 'lion, king of beasts, Panthera leo',
             292: 'tiger, Panthera tigris',
             293: 'cheetah, chetah, Acinonyx jubatus',
             294: 'brown bear, bruin, Ursus arctos',
             295: 'American black bear, black bear, Ursus americanus, Euarctos americanus',
             296: 'ice bear, polar bear, Ursus Maritimus, Thalarctos maritimus',
             297: 'sloth bear, Melursus ursinus, Ursus ursinus',
             298: 'mongoose',
             299: 'meerkat, mierkat',
             300: 'tiger beetle',
             301: 'ladybug, ladybeetle, lady beetle, ladybird, ladybird beetle',
             302: 'ground beetle, carabid beetle',
             303: 'long-horned beetle, longicorn, longicorn beetle',
             304: 'leaf beetle, chrysomelid',
             305: 'dung beetle',
             306: 'rhinoceros beetle',
             307: 'weevil',
             308: 'fly',
             309: 'bee',
             310: 'ant, emmet, pismire',
             311: 'grasshopper, hopper',
             312: 'cricket',
             313: 'walking stick, walkingstick, stick insect',
             314: 'cockroach, roach',
             315: 'mantis, mantid',
             316: 'cicada, cicala',
             317: 'leafhopper',
             318: 'lacewing, lacewing fly',
             319: "dragonfly, darning needle, devil's darning needle, sewing needle, snake feeder, snake doctor, mosquito hawk, skeeter hawk",
             320: 'damselfly',
             321: 'admiral',
             322: 'ringlet, ringlet butterfly',
             323: 'monarch, monarch butterfly, milkweed butterfly, Danaus plexippus',
             324: 'cabbage butterfly',
             325: 'sulphur butterfly, sulfur butterfly',
             326: 'lycaenid, lycaenid butterfly',
             327: 'starfish, sea star',
             328: 'sea urchin',
             329: 'sea cucumber, holothurian',
             330: 'wood rabbit, cottontail, cottontail rabbit',
             331: 'hare',
             332: 'Angora, Angora rabbit',
             333: 'hamster',
             334: 'porcupine, hedgehog',
             335: 'fox squirrel, eastern fox squirrel, Sciurus niger',
             336: 'marmot',
             337: 'beaver',
             338: 'guinea pig, Cavia cobaya',
             339: 'sorrel',
             340: 'zebra',
             341: 'hog, pig, grunter, squealer, Sus scrofa',
             342: 'wild boar, boar, Sus scrofa',
             343: 'warthog',
             344: 'hippopotamus, hippo, river horse, Hippopotamus amphibius',
             345: 'ox',
             346: 'water buffalo, water ox, Asiatic buffalo, Bubalus bubalis',
             347: 'bison',
             348: 'ram, tup',
             349: 'bighorn, bighorn sheep, cimarron, Rocky Mountain bighorn, Rocky Mountain sheep, Ovis canadensis',
             350: 'ibex, Capra ibex',
             351: 'hartebeest',
             352: 'impala, Aepyceros melampus',
             353: 'gazelle',
             354: 'Arabian camel, dromedary, Camelus dromedarius',
             355: 'llama',
             356: 'weasel',
             357: 'mink',
             358: 'polecat, fitch, foulmart, foumart, Mustela putorius',
             359: 'black-footed ferret, ferret, Mustela nigripes',
             360: 'otter',
             361: 'skunk, polecat, wood pussy',
             362: 'badger',
             363: 'armadillo',
             364: 'three-toed sloth, ai, Bradypus tridactylus',
             365: 'orangutan, orang, orangutang, Pongo pygmaeus',
             366: 'gorilla, Gorilla gorilla',
             367: 'chimpanzee, chimp, Pan troglodytes',
             368: 'gibbon, Hylobates lar',
             369: 'siamang, Hylobates syndactylus, Symphalangus syndactylus',
             370: 'guenon, guenon monkey',
             371: 'patas, hussar monkey, Erythrocebus patas',
             372: 'baboon',
             373: 'macaque',
             374: 'langur',
             375: 'colobus, colobus monkey',
             376: 'proboscis monkey, Nasalis larvatus',
             377: 'marmoset',
             378: 'capuchin, ringtail, Cebus capucinus',
             379: 'howler monkey, howler',
             380: 'titi, titi monkey',
             381: 'spider monkey, Ateles geoffroyi',
             382: 'squirrel monkey, Saimiri sciureus',
             383: 'Madagascar cat, ring-tailed lemur, Lemur catta',
             384: 'indri, indris, Indri indri, Indri brevicaudatus',
             385: 'Indian elephant, Elephas maximus',
             386: 'African elephant, Loxodonta africana',
             387: 'lesser panda, red panda, panda, bear cat, cat bear, Ailurus fulgens',
             388: 'giant panda, panda, panda bear, coon bear, Ailuropoda melanoleuca',
             389: 'barracouta, snoek',
             390: 'eel',
             391: 'coho, cohoe, coho salmon, blue jack, silver salmon, Oncorhynchus kisutch',
             392: 'rock beauty, Holocanthus tricolor',
             393: 'anemone fish',
             394: 'sturgeon',
             395: 'gar, garfish, garpike, billfish, Lepisosteus osseus',
             396: 'lionfish',
             397: 'puffer, pufferfish, blowfish, globefish',
             398: 'abacus',
             399: 'abaya',
             400: "academic gown, academic robe, judge's robe",
             401: 'accordion, piano accordion, squeeze box',
             402: 'acoustic guitar',
             403: 'aircraft carrier, carrier, flattop, attack aircraft carrier',
             404: 'airliner',
             405: 'airship, dirigible',
             406: 'altar',
             407: 'ambulance',
             408: 'amphibian, amphibious vehicle',
             409: 'analog clock',
             410: 'apiary, bee house',
             411: 'apron',
             412: 'ashcan, trash can, garbage can, wastebin, ash bin, ash-bin, ashbin, dustbin, trash barrel, trash bin',
             413: 'assault rifle, assault gun',
             414: 'backpack, back pack, knapsack, packsack, rucksack, haversack',
             415: 'bakery, bakeshop, bakehouse',
             416: 'balance beam, beam',
             417: 'balloon',
             418: 'ballpoint, ballpoint pen, ballpen, Biro',
             419: 'Band Aid',
             420: 'banjo',
             421: 'bannister, banister, balustrade, balusters, handrail',
             422: 'barbell',
             423: 'barber chair',
             424: 'barbershop',
             425: 'barn',
             426: 'barometer',
             427: 'barrel, cask',
             428: 'barrow, garden cart, lawn cart, wheelbarrow',
             429: 'baseball',
             430: 'basketball',
             431: 'bassinet',
             432: 'bassoon',
             433: 'bathing cap, swimming cap',
             434: 'bath towel',
             435: 'bathtub, bathing tub, bath, tub',
             436: 'beach wagon, station wagon, wagon, estate car, beach waggon, station waggon, waggon',
             437: 'beacon, lighthouse, beacon light, pharos',
             438: 'beaker',
             439: 'bearskin, busby, shako',
             440: 'beer bottle',
             441: 'beer glass',
             442: 'bell cote, bell cot',
             443: 'bib',
             444: 'bicycle-built-for-two, tandem bicycle, tandem',
             445: 'bikini, two-piece',
             446: 'binder, ring-binder',
             447: 'binoculars, field glasses, opera glasses',
             448: 'birdhouse',
             449: 'boathouse',
             450: 'bobsled, bobsleigh, bob',
             451: 'bolo tie, bolo, bola tie, bola',
             452: 'bonnet, poke bonnet',
             453: 'bookcase',
             454: 'bookshop, bookstore, bookstall',
             455: 'bottlecap',
             456: 'bow',
             457: 'bow tie, bow-tie, bowtie',
             458: 'brass, memorial tablet, plaque',
             459: 'brassiere, bra, bandeau',
             460: 'breakwater, groin, groyne, mole, bulwark, seawall, jetty',
             461: 'breastplate, aegis, egis',
             462: 'broom',
             463: 'bucket, pail',
             464: 'buckle',
             465: 'bulletproof vest',
             466: 'bullet train, bullet',
             467: 'butcher shop, meat market',
             468: 'cab, hack, taxi, taxicab',
             469: 'caldron, cauldron',
             470: 'candle, taper, wax light',
             471: 'cannon',
             472: 'canoe',
             473: 'can opener, tin opener',
             474: 'cardigan',
             475: 'car mirror',
             476: 'carousel, carrousel, merry-go-round, roundabout, whirligig',
             477: "carpenter's kit, tool kit",
             478: 'carton',
             479: 'car wheel',
             480: 'cash machine, cash dispenser, automated teller machine, automatic teller machine, automated teller, automatic teller, ATM',
             481: 'cassette',
             482: 'cassette player',
             483: 'castle',
             484: 'catamaran',
             485: 'CD player',
             486: 'cello, violoncello',
             487: 'cellular telephone, cellular phone, cellphone, cell, mobile phone',
             488: 'chain',
             489: 'chainlink fence',
             490: 'chain mail, ring mail, mail, chain armor, chain armour, ring armor, ring armour',
             491: 'chain saw, chainsaw',
             492: 'chest',
             493: 'chiffonier, commode',
             494: 'chime, bell, gong',
             495: 'china cabinet, china closet',
             496: 'Christmas stocking',
             497: 'church, church building',
             498: 'cinema, movie theater, movie theatre, movie house, picture palace',
             499: 'cleaver, meat cleaver, chopper',
             500: 'cliff dwelling',
             501: 'cloak',
             502: 'clog, geta, patten, sabot',
             503: 'cocktail shaker',
             504: 'coffee mug',
             505: 'coffeepot',
             506: 'coil, spiral, volute, whorl, helix',
             507: 'combination lock',
             508: 'computer keyboard, keypad',
             509: 'confectionery, confectionary, candy store',
             510: 'container ship, containership, container vessel',
             511: 'convertible',
             512: 'corkscrew, bottle screw',
             513: 'cornet, horn, trumpet, trump',
             514: 'cowboy boot',
             515: 'cowboy hat, ten-gallon hat',
             516: 'cradle',
             517: 'crane',
             518: 'crash helmet',
             519: 'crate',
             520: 'crib, cot',
             521: 'Crock Pot',
             522: 'croquet ball',
             523: 'crutch',
             524: 'cuirass',
             525: 'dam, dike, dyke',
             526: 'desk',
             527: 'desktop computer',
             528: 'dial telephone, dial phone',
             529: 'diaper, nappy, napkin',
             530: 'digital clock',
             531: 'digital watch',
             532: 'dining table, board',
             533: 'dishrag, dishcloth',
             534: 'dishwasher, dish washer, dishwashing machine',
             535: 'disk brake, disc brake',
             536: 'dock, dockage, docking facility',
             537: 'dogsled, dog sled, dog sleigh',
             538: 'dome',
             539: 'doormat, welcome mat',
             540: 'drilling platform, offshore rig',
             541: 'drum, membranophone, tympan',
             542: 'drumstick',
             543: 'dumbbell',
             544: 'Dutch oven',
             545: 'electric fan, blower',
             546: 'electric guitar',
             547: 'electric locomotive',
             548: 'entertainment center',
             549: 'envelope',
             550: 'espresso maker',
             551: 'face powder',
             552: 'feather boa, boa',
             553: 'file, file cabinet, filing cabinet',
             554: 'fireboat',
             555: 'fire engine, fire truck',
             556: 'fire screen, fireguard',
             557: 'flagpole, flagstaff',
             558: 'flute, transverse flute',
             559: 'folding chair',
             560: 'football helmet',
             561: 'forklift',
             562: 'fountain',
             563: 'fountain pen',
             564: 'four-poster',
             565: 'freight car',
             566: 'French horn, horn',
             567: 'frying pan, frypan, skillet',
             568: 'fur coat',
             569: 'garbage truck, dustcart',
             570: 'gasmask, respirator, gas helmet',
             571: 'gas pump, gasoline pump, petrol pump, island dispenser',
             572: 'goblet',
             573: 'go-kart',
             574: 'golf ball',
             575: 'golfcart, golf cart',
             576: 'gondola',
             577: 'gong, tam-tam',
             578: 'gown',
             579: 'grand piano, grand',
             580: 'greenhouse, nursery, glasshouse',
             581: 'grille, radiator grille',
             582: 'grocery store, grocery, food market, market',
             583: 'guillotine',
             584: 'hair slide',
             585: 'hair spray',
             586: 'half track',
             587: 'hammer',
             588: 'hamper',
             589: 'hand blower, blow dryer, blow drier, hair dryer, hair drier',
             590: 'hand-held computer, hand-held microcomputer',
             591: 'handkerchief, hankie, hanky, hankey',
             592: 'hard disc, hard disk, fixed disk',
             593: 'harmonica, mouth organ, harp, mouth harp',
             594: 'harp',
             595: 'harvester, reaper',
             596: 'hatchet',
             597: 'holster',
             598: 'home theater, home theatre',
             599: 'honeycomb',
             600: 'hook, claw',
             601: 'hoopskirt, crinoline',
             602: 'horizontal bar, high bar',
             603: 'horse cart, horse-cart',
             604: 'hourglass',
             605: 'iPod',
             606: 'iron, smoothing iron',
             607: "jack-o'-lantern",
             608: 'jean, blue jean, denim',
             609: 'jeep, landrover',
             610: 'jersey, T-shirt, tee shirt',
             611: 'jigsaw puzzle',
             612: 'jinrikisha, ricksha, rickshaw',
             613: 'joystick',
             614: 'kimono',
             615: 'knee pad',
             616: 'knot',
             617: 'lab coat, laboratory coat',
             618: 'ladle',
             619: 'lampshade, lamp shade',
             620: 'laptop, laptop computer',
             621: 'lawn mower, mower',
             622: 'lens cap, lens cover',
             623: 'letter opener, paper knife, paperknife',
             624: 'library',
             625: 'lifeboat',
             626: 'lighter, light, igniter, ignitor',
             627: 'limousine, limo',
             628: 'liner, ocean liner',
             629: 'lipstick, lip rouge',
             630: 'Loafer',
             631: 'lotion',
             632: 'loudspeaker, speaker, speaker unit, loudspeaker system, speaker system',
             633: "loupe, jeweler's loupe",
             634: 'lumbermill, sawmill',
             635: 'magnetic compass',
             636: 'mailbag, postbag',
             637: 'mailbox, letter box',
             638: 'maillot',
             639: 'maillot, tank suit',
             640: 'manhole cover',
             641: 'maraca',
             642: 'marimba, xylophone',
             643: 'mask',
             644: 'matchstick',
             645: 'maypole',
             646: 'maze, labyrinth',
             647: 'measuring cup',
             648: 'medicine chest, medicine cabinet',
             649: 'megalith, megalithic structure',
             650: 'microphone, mike',
             651: 'microwave, microwave oven',
             652: 'military uniform',
             653: 'milk can',
             654: 'minibus',
             655: 'miniskirt, mini',
             656: 'minivan',
             657: 'missile',
             658: 'mitten',
             659: 'mixing bowl',
             660: 'mobile home, manufactured home',
             661: 'Model T',
             662: 'modem',
             663: 'monastery',
             664: 'monitor',
             665: 'moped',
             666: 'mortar',
             667: 'mortarboard',
             668: 'mosque',
             669: 'mosquito net',
             670: 'motor scooter, scooter',
             671: 'mountain bike, all-terrain bike, off-roader',
             672: 'mountain tent',
             673: 'mouse, computer mouse',
             674: 'mousetrap',
             675: 'moving van',
             676: 'muzzle',
             677: 'nail',
             678: 'neck brace',
             679: 'necklace',
             680: 'nipple',
             681: 'notebook, notebook computer',
             682: 'obelisk',
             683: 'oboe, hautboy, hautbois',
             684: 'ocarina, sweet potato',
             685: 'odometer, hodometer, mileometer, milometer',
             686: 'oil filter',
             687: 'organ, pipe organ',
             688: 'oscilloscope, scope, cathode-ray oscilloscope, CRO',
             689: 'overskirt',
             690: 'oxcart',
             691: 'oxygen mask',
             692: 'packet',
             693: 'paddle, boat paddle',
             694: 'paddlewheel, paddle wheel',
             695: 'padlock',
             696: 'paintbrush',
             697: "pajama, pyjama, pj's, jammies",
             698: 'palace',
             699: 'panpipe, pandean pipe, syrinx',
             700: 'paper towel',
             701: 'parachute, chute',
             702: 'parallel bars, bars',
             703: 'park bench',
             704: 'parking meter',
             705: 'passenger car, coach, carriage',
             706: 'patio, terrace',
             707: 'pay-phone, pay-station',
             708: 'pedestal, plinth, footstall',
             709: 'pencil box, pencil case',
             710: 'pencil sharpener',
             711: 'perfume, essence',
             712: 'Petri dish',
             713: 'photocopier',
             714: 'pick, plectrum, plectron',
             715: 'pickelhaube',
             716: 'picket fence, paling',
             717: 'pickup, pickup truck',
             718: 'pier',
             719: 'piggy bank, penny bank',
             720: 'pill bottle',
             721: 'pillow',
             722: 'ping-pong ball',
             723: 'pinwheel',
             724: 'pirate, pirate ship',
             725: 'pitcher, ewer',
             726: "plane, carpenter's plane, woodworking plane",
             727: 'planetarium',
             728: 'plastic bag',
             729: 'plate rack',
             730: 'plow, plough',
             731: "plunger, plumber's helper",
             732: 'Polaroid camera, Polaroid Land camera',
             733: 'pole',
             734: 'police van, police wagon, paddy wagon, patrol wagon, wagon, black Maria',
             735: 'poncho',
             736: 'pool table, billiard table, snooker table',
             737: 'pop bottle, soda bottle',
             738: 'pot, flowerpot',
             739: "potter's wheel",
             740: 'power drill',
             741: 'prayer rug, prayer mat',
             742: 'printer',
             743: 'prison, prison house',
             744: 'projectile, missile',
             745: 'projector',
             746: 'puck, hockey puck',
             747: 'punching bag, punch bag, punching ball, punchball',
             748: 'purse',
             749: 'quill, quill pen',
             750: 'quilt, comforter, comfort, puff',
             751: 'racer, race car, racing car',
             752: 'racket, racquet',
             753: 'radiator',
             754: 'radio, wireless',
             755: 'radio telescope, radio reflector',
             756: 'rain barrel',
             757: 'recreational vehicle, RV, R.V.',
             758: 'reel',
             759: 'reflex camera',
             760: 'refrigerator, icebox',
             761: 'remote control, remote',
             762: 'restaurant, eating house, eating place, eatery',
             763: 'revolver, six-gun, six-shooter',
             764: 'rifle',
             765: 'rocking chair, rocker',
             766: 'rotisserie',
             767: 'rubber eraser, rubber, pencil eraser',
             768: 'rugby ball',
             769: 'rule, ruler',
             770: 'running shoe',
             771: 'safe',
             772: 'safety pin',
             773: 'saltshaker, salt shaker',
             774: 'sandal',
             775: 'sarong',
             776: 'sax, saxophone',
             777: 'scabbard',
             778: 'scale, weighing machine',
             779: 'school bus',
             780: 'schooner',
             781: 'scoreboard',
             782: 'screen, CRT screen',
             783: 'screw',
             784: 'screwdriver',
             785: 'seat belt, seatbelt',
             786: 'sewing machine',
             787: 'shield, buckler',
             788: 'shoe shop, shoe-shop, shoe store',
             789: 'shoji',
             790: 'shopping basket',
             791: 'shopping cart',
             792: 'shovel',
             793: 'shower cap',
             794: 'shower curtain',
             795: 'ski',
             796: 'ski mask',
             797: 'sleeping bag',
             798: 'slide rule, slipstick',
             799: 'sliding door',
             800: 'slot, one-armed bandit',
             801: 'snorkel',
             802: 'snowmobile',
             803: 'snowplow, snowplough',
             804: 'soap dispenser',
             805: 'soccer ball',
             806: 'sock',
             807: 'solar dish, solar collector, solar furnace',
             808: 'sombrero',
             809: 'soup bowl',
             810: 'space bar',
             811: 'space heater',
             812: 'space shuttle',
             813: 'spatula',
             814: 'speedboat',
             815: "spider web, spider's web",
             816: 'spindle',
             817: 'sports car, sport car',
             818: 'spotlight, spot',
             819: 'stage',
             820: 'steam locomotive',
             821: 'steel arch bridge',
             822: 'steel drum',
             823: 'stethoscope',
             824: 'stole',
             825: 'stone wall',
             826: 'stopwatch, stop watch',
             827: 'stove',
             828: 'strainer',
             829: 'streetcar, tram, tramcar, trolley, trolley car',
             830: 'stretcher',
             831: 'studio couch, day bed',
             832: 'stupa, tope',
             833: 'submarine, pigboat, sub, U-boat',
             834: 'suit, suit of clothes',
             835: 'sundial',
             836: 'sunglass',
             837: 'sunglasses, dark glasses, shades',
             838: 'sunscreen, sunblock, sun blocker',
             839: 'suspension bridge',
             840: 'swab, swob, mop',
             841: 'sweatshirt',
             842: 'swimming trunks, bathing trunks',
             843: 'swing',
             844: 'switch, electric switch, electrical switch',
             845: 'syringe',
             846: 'table lamp',
             847: 'tank, army tank, armored combat vehicle, armoured combat vehicle',
             848: 'tape player',
             849: 'teapot',
             850: 'teddy, teddy bear',
             851: 'television, television system',
             852: 'tennis ball',
             853: 'thatch, thatched roof',
             854: 'theater curtain, theatre curtain',
             855: 'thimble',
             856: 'thresher, thrasher, threshing machine',
             857: 'throne',
             858: 'tile roof',
             859: 'toaster',
             860: 'tobacco shop, tobacconist shop, tobacconist',
             861: 'toilet seat',
             862: 'torch',
             863: 'totem pole',
             864: 'tow truck, tow car, wrecker',
             865: 'toyshop',
             866: 'tractor',
             867: 'trailer truck, tractor trailer, trucking rig, rig, articulated lorry, semi',
             868: 'tray',
             869: 'trench coat',
             870: 'tricycle, trike, velocipede',
             871: 'trimaran',
             872: 'tripod',
             873: 'triumphal arch',
             874: 'trolleybus, trolley coach, trackless trolley',
             875: 'trombone',
             876: 'tub, vat',
             877: 'turnstile',
             878: 'typewriter keyboard',
             879: 'umbrella',
             880: 'unicycle, monocycle',
             881: 'upright, upright piano',
             882: 'vacuum, vacuum cleaner',
             883: 'vase',
             884: 'vault',
             885: 'velvet',
             886: 'vending machine',
             887: 'vestment',
             888: 'viaduct',
             889: 'violin, fiddle',
             890: 'volleyball',
             891: 'waffle iron',
             892: 'wall clock',
             893: 'wallet, billfold, notecase, pocketbook',
             894: 'wardrobe, closet, press',
             895: 'warplane, military plane',
             896: 'washbasin, handbasin, washbowl, lavabo, wash-hand basin',
             897: 'washer, automatic washer, washing machine',
             898: 'water bottle',
             899: 'water jug',
             900: 'water tower',
             901: 'whiskey jug',
             902: 'whistle',
             903: 'wig',
             904: 'window screen',
             905: 'window shade',
             906: 'Windsor tie',
             907: 'wine bottle',
             908: 'wing',
             909: 'wok',
             910: 'wooden spoon',
             911: 'wool, woolen, woollen',
             912: 'worm fence, snake fence, snake-rail fence, Virginia fence',
             913: 'wreck',
             914: 'yawl',
             915: 'yurt',
             916: 'web site, website, internet site, site',
             917: 'comic book',
             918: 'crossword puzzle, crossword',
             919: 'street sign',
             920: 'traffic light, traffic signal, stoplight',
             921: 'book jacket, dust cover, dust jacket, dust wrapper',
             922: 'menu',
             923: 'plate',
             924: 'guacamole',
             925: 'consomme',
             926: 'hot pot, hotpot',
             927: 'trifle',
             928: 'ice cream, icecream',
             929: 'ice lolly, lolly, lollipop, popsicle',
             930: 'French loaf',
             931: 'bagel, beigel',
             932: 'pretzel',
             933: 'cheeseburger',
             934: 'hotdog, hot dog, red hot',
             935: 'mashed potato',
             936: 'head cabbage',
             937: 'broccoli',
             938: 'cauliflower',
             939: 'zucchini, courgette',
             940: 'spaghetti squash',
             941: 'acorn squash',
             942: 'butternut squash',
             943: 'cucumber, cuke',
             944: 'artichoke, globe artichoke',
             945: 'bell pepper',
             946: 'cardoon',
             947: 'mushroom',
             948: 'Granny Smith',
             949: 'strawberry',
             950: 'orange',
             951: 'lemon',
             952: 'fig',
             953: 'pineapple, ananas',
             954: 'banana',
             955: 'jackfruit, jak, jack',
             956: 'custard apple',
             957: 'pomegranate',
             958: 'hay',
             959: 'carbonara',
             960: 'chocolate sauce, chocolate syrup',
             961: 'dough',
             962: 'meat loaf, meatloaf',
             963: 'pizza, pizza pie',
             964: 'potpie',
             965: 'burrito',
             966: 'red wine',
             967: 'espresso',
             968: 'cup',
             969: 'eggnog',
             970: 'alp',
             971: 'bubble',
             972: 'cliff, drop, drop-off',
             973: 'coral reef',
             974: 'geyser',
             975: 'lakeside, lakeshore',
             976: 'promontory, headland, head, foreland',
             977: 'sandbar, sand bar',
             978: 'seashore, coast, seacoast, sea-coast',
             979: 'valley, vale',
             980: 'volcano',
             981: 'ballplayer, baseball player',
             982: 'groom, bridegroom',
             983: 'scuba diver',
             984: 'rapeseed',
             985: 'daisy',
             986: "yellow lady's slipper, yellow lady-slipper, Cypripedium calceolus, Cypripedium parviflorum",
             987: 'corn',
             988: 'acorn',
             989: 'hip, rose hip, rosehip',
             990: 'buckeye, horse chestnut, conker',
             991: 'coral fungus',
             992: 'agaric',
             993: 'gyromitra',
             994: 'stinkhorn, carrion fungus',
             995: 'earthstar',
             996: 'hen-of-the-woods, hen of the woods, Polyporus frondosus, Grifola frondosa',
             997: 'bolete',
             998: 'ear, spike, capitulum',
             999: 'toilet tissue, toilet paper, bathroom tissue'}

cifar_label = {0:  "apple",
               1:  "aquarium_fish",
               2:  "baby",
               3:  "bear",
               4:  "beaver",
               5:  "bed",
               6:  "bee",
               7:  "beetle",
               8:  "bicycle",
               9:  "bottle",
               10: "bowl",
               11: "boy",
               12: "bridge",
               13: "bus",
               14: "butterfly",
               15: "camel",
               16: "can",
               17: "castle",
               18: "caterpillar",
               19: "cattle",
               20: "chair",
               21: "chimpanzee",
               22: "clock",
               23: "cloud",
               24: "cockroach",
               25: "couch",
               26: "cra",
               27: "crocodile",
               28: "cup",
               29: "dinosaur",
               30: "dolphin",
               31: "elephant",
               32: "flatfish",
               33: "forest",
               34: "fox",
               35: "girl",
               36: "hamster",
               37: "house",
               38: "kangaroo",
               39: "keyboard",
               40: "lamp",
               41: "lawn_mower",
               42: "leopard",
               43: "lion",
               44: "lizard",
               45: "lobster",
               46: "man",
               47: "maple_tree",
               48: "motorcycle",
               49: "mountain",
               50: "mouse",
               51: "mushroom",
               52: "oak_tree",
               53: "orange",
               54: "orchid",
               55: "otter",
               56: "palm_tree",
               57: "pear",
               58: "pickup_truck",
               59: "pine_tree",
               60: "plain",
               61: "plate",
               62: "poppy",
               63: "porcupine",
               64: "possum",
               65: "rabbit",
               66: "raccoon",
               67: "ray",
               68: "road",
               69: "rocket",
               70: "rose",
               71: "sea",
               72: "seal",
               73: "shark",
               74: "shrew",
               75: "skunk",
               76: "skyscraper",
               77: "snail",
               78: "snake",
               79: "spider",
               80: "squirrel",
               81: "streetcar",
               82: "sunflower",
               83: "sweet_pepper",
               84: "table",
               85: "tank",
               86: "telephone",
               87: "television",
               88: "tiger",
               89: "tractor",
               90: "train",
               91: "trout",
               92: "tulip",
               93: "turtle",
               94: "wardrobe",
               95: "whale",
               96: "willow_tree",
               97: "wolf",
               98: "woman",
               99: "worm"}


from pytorch_grad_cam import GradCAM, \
    ScoreCAM, \
    GradCAMPlusPlus, \
    AblationCAM, \
    XGradCAM, \
    EigenCAM, \
    EigenGradCAM, \
    LayerCAM, \
    FullGrad

def reshape_swin(data, height=14, width=14):
    result = data.reshape(data.size(0), height, width, data.size(2))

    result = result.transpose(2, 3).transpose(1, 2)

    return result

def reshape_vit(data, height=14, width=14):
    result = data[:, 1:  , :].reshape(data.size(0), height, width, data.size(2))

    # Bring the channels to the first dimension,
    # like in CNNs.
    result = result.transpose(2, 3).transpose(1, 2)
    return result


methods_grad = \
    {"gradcam": GradCAM,
     "scorecam": ScoreCAM,
     "gradcam++": GradCAMPlusPlus,
     "ablationcam": AblationCAM,
     "xgradcam": XGradCAM,
     "eigencam": EigenCAM,
     "eigengradcam": EigenGradCAM,
     "layercam": LayerCAM,
     "fullgrad": FullGrad}












def get_update_param(model):
    return [k for k, v in model.named_parameters() if v.requires_grad == True]

def shape(param):
    return f"({', '.join(map(str, param.shape))})"

def analysis_mode():
    from joonmyung.draw import drawImgPlot, drawLinePlot, drawHeatmap, drawBarChart
    from joonmyung.utils import to_np
    print(get_update_param)
    print(drawImgPlot)

if __name__ == '__main__':
    analysis_mode()



from collections import OrderedDict

from joonmyung.utils import isDir
from timm import create_model
import torch
import os

class JModel():
    def __init__(self, num_classes = None, model_path= None, device="cuda", p=False):
        # Pretrained_Model
        self.num_classes = num_classes

        if model_path:
            self.model_path = os.path.join(model_path, "checkpoint_{}.pth")
        if p and model_path:
            print("file list : ", sorted(os.listdir(model_path), reverse=True))
        self.device = device

    def load_state_dict(self, model, state_dict):
        state_dict = OrderedDict((k.replace("module.", ""), v) for k, v in state_dict.items())
        model.load_state_dict(state_dict)


    def getModel(self, model_type=0, model_name ="deit_tiny", **kwargs):

        if model_type == 0:
            model = create_model(model_name, pretrained=True, num_classes=self.num_classes, in_chans=3, global_pool=None, scriptable=False)
        elif model_type == 1:
            model = torch.hub.load('facebookresearch/deit:main', model_name, pretrained=True)
        else:
            raise ValueError
        model.eval()

        return model.to(self.device)




from fvcore.nn import FlopCountAnalysis, flop_count_table
from torchprofile import profile_macs
from typing import Tuple
from thop import profile
from tqdm import tqdm
import torch
import time

def thop(model, size, *kwargs,
         round_num=1, eval = True, device="cuda"):
    if eval: model.eval().to(device)
    input = torch.randn(size, device=device)
    macs, params = profile(model, inputs=(input, *kwargs))
    macs, params = macs / 1000000000, params / 1000000

    print(f"thop macs/params : {macs}/{params}")

    return round(macs, round_num), round(params, round_num)

def numel(model,
          round_num=1):
    params = sum([p.numel() for p in model.parameters() if p.requires_grad]) / 1000000
    print(f"numel params : {params}")
    return round(params, round_num)

@torch.no_grad()
def flops(model, size, round_num=1, eval=True, fp16=False, device="cuda", **kwargs):
    if eval: model.eval()
    with torch.cuda.amp.autocast(enabled=fp16):
        inputs = torch.randn(size, device=device, requires_grad=True)
        with torch.no_grad():
            flops = FlopCountAnalysis(model, (inputs, *kwargs))
            flops_num = flops.total() / 1000000000

    print(flop_count_table(flops))
    print(f"fvcore flops : {flops_num}")

    return round(flops_num, round_num)


def get_macs(model, size,
             eval = True, round_num=1, device="cuda"):
    if eval: model.eval()
    inputs = torch.randn(size, device=device, requires_grad=True)
    macs = profile_macs(model, inputs)
    print(f"torchprofile MACS : {macs}")

    return round(macs, round_num)

def benchmark(
    model: torch.nn.Module,
    device: torch.device = 0,
    input_size: Tuple[int] = (3, 224, 224),
    batch_size: int = 1024,
    runs: int = 40,
    throw_out: float = 0.25,
    use_fp16: bool = False,
    verbose: bool = False,
    **kwargs
) -> float:
    """
    Benchmark the given model with random inputs at the given batch size.

    Args:
     - model: the module to benchmark
     - device: the device to use for benchmarking
     - input_size: the input size to pass to the model (channels, h, w)
     - batch_size: the batch size to use for evaluation
     - runs: the number of total runs to do
     - throw_out: the percentage of runs to throw out at the start of testing
     - use_fp16: whether or not to benchmark with float16 and autocast
     - verbose: whether or not to use tqdm to print progress / print throughput at end

    Returns:
     - the throughput measured in images / second
    """

    if not isinstance(device, torch.device):
        device = torch.device(device)
    is_cuda = torch.device(device).type == "cuda"

    model = model.eval().to(device)
    input = torch.rand(batch_size, *input_size, device=device)
    if use_fp16:
        input = input.half()

    warm_up = int(runs * throw_out)
    total = 0
    start = time.time()

    with torch.autocast(device.type, enabled=use_fp16):
        with torch.no_grad():
            for i in tqdm(range(runs), disable=not verbose, desc="Benchmarking"):
                if i == warm_up:
                    if is_cuda:
                        torch.cuda.synchronize()
                    total = 0
                    start = time.time()

                model(input, **kwargs)
                total += batch_size


    if is_cuda:
        torch.cuda.synchronize()

    end = time.time()
    elapsed = end - start

    throughput = total / elapsed

    if verbose:
        print(f"Throughput: {throughput:.2f} im/s")

    return round(throughput)



def accuracy(output, target, topk=(1,)):
    """Computes the accuracy over the k top predictions for the specified values of k"""
    maxk = max(topk)
    _, pred = output.topk(maxk, 1, True, True)
    batch_size = target.size(0)
    pred = pred.t()
    correct = pred.eq(target.reshape(1, -1).expand_as(pred))
    return [correct[:k].reshape(-1).float().sum(0) * 100. / batch_size for k in topk]

def targetPred(output, target, topk=5):
    _, pred = output.topk(topk, 1, True, True)
    TP = torch.cat([target.unsqueeze(-1), pred], dim=1)
    return TP


from joonmyung.meta_data.label import imnet_label, cifar_label
import torch.nn.functional as F
import pandas as pd
import numpy as np
import getpass
import socket
import torch
import os

def data2path(dataset,
              conference="", wandb_version="", wandb_name="", hub_num = 1):

    hostname = socket.gethostname()
    server   = hostname if "mlv" in hostname \
                else "kakao" if "dakao" in hostname \
                    else "kisti_"+hostname
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    print(server, " : ", hostname, " : ", getpass.getuser())
    print("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@")
    if "kakao" in server:
        data_path  = "/data/opensets"
        output_dir = f"/data/project/rw/{getpass.getuser()}/conference"
    elif "mlv" in server:
        data_path  = f"/hub_data{hub_num}/{getpass.getuser()}/data"
        output_dir = f"/hub_data{hub_num}/{getpass.getuser()}/conference"
    elif "kisti" in server:
        data_path  = f"/scratch/{getpass.getuser()}/data"
        output_dir = f"/scratch/{getpass.getuser()}/result"
    else:
        raise ValueError

    if dataset in ["imagenet", "IMNET"]:
        data_path = os.path.join(data_path, "imagenet") if "kakao" not in server else os.path.join(data_path, "imagenet-pytorch")
        num_classes = 1000
    else:
        raise ValueError

    output_dir    = os.path.join(output_dir, conference, wandb_version, wandb_name)

    return data_path, num_classes, output_dir, server

def get_label(key, d_name ="imagenet"):
    d_name = d_name.lower()
    if d_name in ["imagenet", "IMNET"] :
        return imnet_label[key]
    elif d_name in ["cifar10", "cifar100"]:
        return cifar_label[key]


def makeSample(shape, min=0, max=1, dataType=int, outputType=np):
    if dataType == int:
        d = np.random.randint(min, max, size=shape)
    elif dataType == float:
        d = np.random.uniform(low=min, high=max, size=shape)
    else:
        raise ValueError

    if outputType == np:
        return d
    elif outputType == pd:
        return pd.DataFrame(d, columns=None)
    elif outputType == torch:
        return torch.from_numpy(d)

def makeAttn(shape, dim=1):
    return F.softmax(torch.randn(shape), dim=dim)

def set_dtype(df, dtypes):
    for c_n, d_t in dtypes.items():
        if c_n in df.columns:
            df[c_n] = df[c_n].astype(d_t)
    return df


import setuptools
from setuptools import find_packages

setuptools.setup(
    name="joonmyung",
    version="1.5.2",
    author="JoonMyung Choi",
    author_email="pizard@korea.ac.kr",
    description="JoonMyung's Library",
    url="https://github.com/pizard/JoonMyung.git",
    license="MIT",
    packages=find_packages(exclude=["playground",
                                    "playground.*",
                                    "99_backup",
                                    "99_backup.*",
                                    "*.egg-info"
                                    "*.egg-info.*"]),
    zip_safe=False,
    install_requires=[
    ]
)

# git add .
# git commit
# git push
# rm -rf ./*.egg-info ./dist/*

# python setup.py sdist; python -m twine upload dist/*
# ID:JoonmyungChoi


import torch.distributed as dist
from pathlib import Path
import numpy as np
import argparse
import zipfile
import torch
import ast
import os

def to_leaf(vs):
    return [v.detach().cpu() for v in vs] if type(vs) == list else vs.detach().cpu()

def to_np(vs):
    if type(vs) == list:
        return [v.detach().cpu().numpy() if type(v) is not np.ndarray else v for v in vs]
    else:
        return vs.detach().cpu().numpy() if type(vs) is not np.ndarray else vs

def to_tensor(vs):
    if type(vs) == list:
        return [torch.Tensor(v) if type(v) is not torch.Tensor else v for v in vs]
    else:
        return torch.Tensor(vs) if type(vs) is not torch.Tensor else vs

def str2list(s):
    v = ast.literal_eval(s.replace(" ", ""))
    if type(v) is not list:
        raise argparse.ArgumentTypeError("Argument \"%s\" is not a list" % (s))
    return v

def str2bool(v):
    if isinstance(v, bool):
        return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')

def time2str(time, type = 0):
    if type == 0:
        return "{:4d}.{:2d}.{:2d} {:2d}:{:2d}:{:2d}".format(time.tm_year, time.tm_mon, time.tm_mday, time.tm_hour, time.tm_min, time.tm_sec)
    else:
        raise ValueError()

def is_main_process():
    return get_rank() == 0

def get_rank():
    if not is_dist_avail_and_initialized():
        return 0
    return dist.get_rank()

def is_dist_avail_and_initialized():
    if not dist.is_available():
        return False
    if not dist.is_initialized():
        return False
    return True


def make_zipfile(src_dir, save_path, enclosing_dir="", exclude_dirs=None, exclude_extensions=None,
                 exclude_dirs_substring=None):
    """make a zip file of root_dir, save it to save_path.
    exclude_paths will be excluded if it is a subdir of root_dir.
    An enclosing_dir is added is specified.
    """
    abs_src = os.path.abspath(src_dir)
    with zipfile.ZipFile(save_path, "w") as zf:
        for dirname, subdirs, files in os.walk(src_dir):
            if exclude_dirs is not None:
                for e_p in exclude_dirs:
                    if e_p in subdirs:
                        subdirs.remove(e_p)
            if exclude_dirs_substring is not None:
                to_rm = []
                for d in subdirs:
                    if exclude_dirs_substring in d:
                        to_rm.append(d)
                for e in to_rm:
                    subdirs.remove(e)
            arcname = os.path.join(enclosing_dir, dirname[len(abs_src) + 1:])
            zf.write(dirname, arcname)
            for filename in files:
                if exclude_extensions is not None:
                    if os.path.splitext(filename)[1] in exclude_extensions:
                        continue  # do not zip it
                absname = os.path.join(dirname, filename)
                arcname = os.path.join(enclosing_dir, absname[len(abs_src) + 1:])
                zf.write(absname, arcname)

    #     # save a copy of the codebase. !!!Do not store heavy file in your codebase when using it.
    #     code_dir = dirname(dirname(dirname(os.path.realpath(__file__))))   # '/data/project/rw/joonmyung/conference/2023CVPR/code_alpro'
    #     code_zip_filename = os.path.join(args.output_dir, "code.zip")
    #     LOGGER.info(f"Saving code from {code_dir} to {code_zip_filename}...")
    #     make_zipfile(code_dir, code_zip_filename,
    #                  enclosing_dir="code",
    #                  exclude_dirs_substring="results",
    #                  exclude_dirs=["__pycache__", "output", "data", "ext"],
    #                  exclude_extensions=[".pyc", ".ipynb", ".swap", ".pt"])
    #     LOGGER.info(f"Saving code done.")

def getDir(path):
    # return [d for d in os.listdir(path) if os.path.isdir(os.path.join(path, d))]
    return [item.name for item in Path(path).iterdir() if item.is_dir()]

def isDir(path):
    return os.path.exists(path)


from flask import Flask, request, render_template
import queue
import threading
import subprocess
from werkzeug.middleware.proxy_fix import ProxyFix

app = Flask(__name__)
process_queue = queue.PriorityQueue()
next_pidx = 0

@app.route('/')
def home():
    return '''
        <!DOCTYPE html>
        <html>
            <head>
                <title>Process Queue</title>
                <style>
                    body {
                        font-family: Arial, sans-serif;
                    }
                    h1 {
                        text-align: center;
                    }
                    form {
                        display: flex;
                        flex-direction: column;
                        width: 50%;
                        margin: 0 auto;
                    }
                    label {
                        margin-bottom: 0.5rem;
                    }
                    input[type="text"],
                    input[type="number"] {
                        padding: 0.5rem;
                        border-radius: 4px;
                        border: 1px solid #ccc;
                        margin-bottom: 1rem;
                    }
                    button[type="submit"] {
                        padding: 0.5rem;
                        border-radius: 4px;
                        background-color: #4CAF50;
                        color: #fff;
                        border: none;
                        cursor: pointer;
                    }
                    button[type="submit"]:hover {
                        background-color: #3e8e41;
                    }
                    a {
                        display: block;
                        margin-top: 2rem;
                        text-align: center;
                    }
                    ul {
                        list-style: none;
                        padding-left: 0;
                    }
                    li {
                        margin-bottom: 1rem;
                        display: flex;
                        justify-content: space-between;
                        align-items: center;
                    }
                    li button[type="submit"] {
                        background-color: #f44336;
                    }
                    li button[type="submit"]:hover {
                        background-color: #c62828;
                    }
                    li form:first-child {
                        margin-right: 1rem;
                    }
                    form:last-of-type {
                        margin-top: 2rem;
                    }
                </style>
            </head>
            <body>
                <h1>Process Queue</h1>
                <form method="POST" action="/enqueue">
                    <label for="process">Process:</label>
                    <input type="text" id="process" name="process" required>
                    <label for="priority">Priority:</label>
                    <input type="number" id="priority" name="priority" required>
                    <button type="submit">Enqueue</button>
                </form>
                <a href="/queue">View Queue</a>
            </body>
        </html>
    '''


# from flask import Flask, request
# import queue
# import threading
# import subprocess

# app = Flask(__name__)
# process_queue = queue.PriorityQueue()
# next_pidx = 0

# @app.route('/')
# def home():
#     return '''
#         <form method="POST" action="/enqueue">
#             <label for="process">Process:</label>
#             <input type="text" id="process" name="process" required>
#             <label for="priority">Priority:</label>
#             <input type="number" id="priority" name="priority" required>
#             <button type="submit">Enqueue</button>
#         </form>
#         <a href="/queue">View Queue</a>
#     '''

@app.route('/enqueue', methods=['POST'])
def enqueue():
    global next_pidx
    process = request.form['process']
    priority = int(request.form['priority'])
    process_queue.put((priority, process, next_pidx))
    next_pidx += 1
    return 'Process enqueued.'

@app.route('/remove', methods=['POST'])
def remove():
    pidx_to_remove = request.form['process_to_remove']
    for priority, process, pidx in list(process_queue.queue):
        if pidx == int(pidx_to_remove):
            process_queue.queue.remove((priority, process, pidx))
    return 'Process removed.'

@app.route('/change_priority', methods=['POST'])
def change_priority():
    pidx_to_change = request.form['process_to_change']
    new_priority = int(request.form['new_priority'])
    for priority, process, pidx in list(process_queue.queue):
        if pidx == int(pidx_to_change):
            process_queue.queue.remove((priority, process, pidx))
            process_queue.put((new_priority, process, pidx))
    return 'Process priority changed.'

# @app.route('/queue')
# def view_queue():
#     items = list(process_queue.queue)
#     if not items:
#         return 'Queue is empty.'
#     else:
#         items_html = '<ul>'
#         for priority, process, pidx in items:
#             items_html += f'''
#                 <li>
#                     {process} (priority: {priority}, process index : {pidx})
#                     <form method="POST" action="/remove" style="display: inline-block;">
#                         <input type="hidden" name="process_to_remove" value="{pidx}">
#                         <button type="submit">Remove</button>
#                     </form>
#                     <form method="POST" action="/change_priority" style="display: inline-block;">
#                         <input type="hidden" name="process_to_change" value="{pidx}">
#                         <label for="new_priority">New priority:</label>
#                         <input type="number" id="new_priority" name="new_priority" required>
#                         <button type="submit">Change priority</button>
#                     </form>
#                 </li>
#             '''
#         items_html += '</ul>'
#         return items_html


@app.route('/queue')
def view_queue():
    items = list(process_queue.queue)
    return render_template('queue.html', items=items)


def execute_processes():
    while True:
        priority, process, pidx = process_queue.get()
        print(f'Executing process: {process} with process index : {pidx}')
        subprocess.run(process, shell=True, stdout=subprocess.DEVNULL)
        process_queue.task_done()
        print(f'{process} done.')

if __name__ == '__main__':
    execution_thread = threading.Thread(target=execute_processes)
    execution_thread.start()
    app.run(host='0.0.0.0', port='5000')



from torchvision.transforms.functional import to_pil_image
from joonmyung.utils import to_leaf, to_tensor
from joonmyung.data import normalization
from matplotlib import pyplot as plt
import torch.utils.data.distributed
import matplotlib as mpl
import torch.nn.parallel
import torch.utils.data
from PIL import Image
from tqdm import tqdm
import seaborn as sns
import pandas as pd
import numpy as np
import torch.optim
import torch
import cv2
import PIL
import os

def drawHeatmap(matrixes, col=1, title=[], fmt=1, p=False,
                vmin=None, vmax=None, xticklabels=False, yticklabels=False,
                linecolor=None, linewidths=0.1, fontsize=30, r=[1,1],
                cmap="Greys", cbar=True, border=False,
                output_dir='./result', save_name=None, show=True,
                vis_x= False, vis_y =  False):
    row = (len(matrixes) - 1) // col + 1
    annot = True if fmt > 0 else False
    if p:
        print("|- Parameter Information")
        print("  |- Data Info (G, H, W)")
        print("    |- G : Graph Num")
        print("    |- H : height data dimension")
        print("    |- W : weidth data dimension")
        print("  |- AXIS Information")
        print("    |- col        : 컬럼 갯수")
        print("    |- row        : {}, col : {}".format(row, col))
        print("    |- height     : {}, width : {}".format(row * 8, col * 8))
        print("    |- title      : 컬럼별 제목")
        print("    |- p          : 정보 출력")
        print()
        print("  |- Graph Information")
        print("    |- vmin/vmax  : 히트맵 최소/최대 값")
        print("    |- linecolor  : black, ...   ")
        print("    |- linewidths : 1.0...   ")
        print("    |- fmt        : 숫자 출력 소숫점 자릿 수")
        print("    |- cmap        : Grey")
        print("    |- cbar        : 오른쪽 바 On/Off")
        print("    |- xticklabels : x축 간격 (False, 1,2,...)")
        print("    |- yticklabels : y축 간격 (False, 1,2,...)")

    if title:
        title = title + list(range(len(title), len(matrixes) - len(title)))
    fig, axes = plt.subplots(nrows=row, ncols=col, squeeze=False)
    fig.set_size_inches(col * 8 * r[1], row * 8 * r[0])
    # fig.set_size_inches(8, 8)
    fig.patch.set_facecolor('white')
    for e, matrix in enumerate(tqdm(matrixes)):
        if type(matrix) == torch.Tensor:
            matrix = matrix.detach().cpu().numpy()
        ax = axes[e // col][e % col]
        res = sns.heatmap(pd.DataFrame(matrix), annot=annot, fmt=".{}f".format(fmt), cmap=cmap
                          , vmin=vmin, vmax=vmax, yticklabels=yticklabels, xticklabels=xticklabels
                          , linewidths=linewidths, linecolor=linecolor, cbar=cbar, annot_kws={"size": fontsize / np.sqrt(len(matrix))}
                          , ax=ax)

        # if border:
        #     for _, spine in res.spines.items():
        #         spine.set_visible(True)
        if not border:
            ax.set_axis_off()
        # if columns:
        #     ax.set_title(columns[c_num] + str(r_num)) if len(columns) == col else ax.set_title(columns[i])
        if not vis_x: ax.xaxis.set_visible(False)
        if not vis_y: ax.yaxis.set_visible(False)

        if title:
            ax.set(title="{} : {}".format(title, e))
    plt.tight_layout()
    if output_dir and save_name:
        if not os.path.exists(output_dir):
            os.makedirs(output_dir, exist_ok=True)
        plt.savefig(os.path.join(output_dir, save_name), transparent = True)
    if show:
        plt.show()


def drawLinePlot(datas, index, col=1, title=[], xlabels=None, ylabels=None, markers=False, columns=None, p=False):
    row = (len(datas) - 1) // col + 1
    title = title + list(range(len(title), len(datas) - len(title)))
    fig, axes = plt.subplots(nrows=row, ncols=col, squeeze=False)
    fig.set_size_inches(col * 8, row * 8)

    if p:
        print("|- Parameter Information")
        print("  |- Data Info (G, D, C)")
        print("    |- G : Graph Num")
        print("    |- D : x data Num (Datas)")
        print("    |- C : y data Num (Column)")
        print("  |- Axis Info")
        print("    |- col   : 컬럼 갯수")
        print("    |- row : {}, col : {}".format(row, col))
        print("    |- height : {}, width : {}".format(row * 8, col * 8))
        print("    |- title : 컬럼별 제목")
        print("    |- p     : 정보 출력")
        print("  |- Graph Info")
        print("    |- vmin/row  : 히트맵 최소/최대 값")
        print("    |- linecolor  : black, ...   ")
        print("    |- linewidths : 1.0...   ")
        print("    |- fmt        : 숫자 출력 소숫점 자릿 수")
        print("    |- cmap        : Grey")
        print("    |- cbar        : 오른쪽 바 On/Off")
        print("    |- xticklabels : x축 간격 (False, 1,2,...)")
        print("    |- yticklabels : y축 간격 (False, 1,2,...)")
        print()

    for e, data in enumerate(datas):
        ax = axes[e // col][e % col]
        d = pd.DataFrame(data, index=index, columns=columns).reset_index()
        d = d.melt(id_vars=["index"], value_vars=columns)
        p = sns.lineplot(x="index", y="value", data=d, hue="variable", markers=markers, ax=ax)
        p.set_xlabel(xlabels, fontsize=20)
        p.set_ylabel(ylabels, fontsize=20)

        ax.set(title=title[e])
    plt.legend(bbox_to_anchor=(1.05, 1.0), loc='upper left')
    plt.tight_layout()
    plt.show()

def drawBarChart(df, x, y, splitColName, col=1, title=[], fmt=1, p=False, c=False, c_sites={}, showfliers=True):
    d2s = df[splitColName].unique()
    d1 = df['d1'].unique()[0]
    d2s = [d2 for d2 in d2s for c_site in c_sites[d1].keys() if c_site in d2]

    row = (len(d2s) - 1) // col + 1

    fig, axes = plt.subplots(nrows=row, ncols=col, squeeze=False)
    fig.set_size_inches(col * 12, row * 12)
    for e, d2 in enumerate(tqdm(d2s)):
        plt.title(d2, fontsize=20)
        ax = axes[e // col][e % col]
        temp = df.loc[df['d2'].isin([d2])]
        if temp["Date_m"].max() != temp["Date_m"].min():
            ind = pd.date_range(temp["Date_m"].min(), temp["Date_m"].max(), freq="M").strftime("%Y-%m")
        else:
            pd.to_datetime(temp["Date_m"].max()).strftime("%Y-%m")
        g = sns.boxplot(x=x, y=y, data=temp, order=ind, ax=ax, showfliers=showfliers)
        g.set(title=d2)
        g.set_xticklabels(ind, rotation=45, fontsize=15)
        g.set(xlabel=None)
        for c_site, c_dates in c_sites[d1].items():
            if c_site in d2:
                for c_date in c_dates:
                    c_ind = (pd.to_datetime(c_date, format='%Y%m%d') - pd.to_datetime(temp["Date_m"].min())).days / 30
                    if c_ind >= 0:
                        g.axvline(c_ind, ls='--', c="red")
    plt.show()

@torch.no_grad()
def saliency(attentions=None, gradients=None, head_fusion="mean",
             discard_ratios = [0.0], data_from="cls", ls_rollout=[], ls_attentive=[],
             reshape=False, device="cpu"):
    # attentions : L * (B, H, h, w), gradients : L * (B, H, h, w)
    if type(discard_ratios) is not list: discard_ratios = [discard_ratios]
    saliencys = 1.
    if attentions:
        attentions = torch.stack(attentions, dim=0)
        saliencys = saliencys * attentions
    if gradients:
        gradients = torch.stack(gradients, dim=0)
        saliencys = saliencys * gradients

    if head_fusion == "mean":
        saliencys = saliencys.mean(axis=2) #
    elif head_fusion == "max":
        saliencys = saliencys.max(axis=2)[0]
    elif head_fusion == "min":
        saliencys = saliencys.min(axis=2)[0]
    elif head_fusion == "median":
        saliencys = saliencys.median(axis=2)[0]

    saliencys = saliencys.to(device)

    _, B, _, T = saliencys.shape
    H = W = int(T ** 0.5)

    rollouts, attentive = None, None
    if ls_rollout:
        rollouts, I = [], torch.eye(T, device=device).unsqueeze(0).expand(B, -1, -1)  # (B, 197, 197)
        for discard_ratio in discard_ratios:
            for start in ls_rollout:
                rollout = I
                for attn in copy.deepcopy(saliencys[start:]):  # (L, B, w, h)
                    # TODO NEED TO CORRECT
                    if discard_ratio:
                        flat = attn.reshape(B, -1)
                        _, indices = flat.topk(int(flat.shape[-1] * discard_ratio), -1, False)
                        indices = indices * (indices != 0)
                        for b in range(B):
                            flat[b, indices[b]] = 0

                    attn = 0.5 * attn + 0.5 * I
                    attn = attn / attn.sum(dim=-1, keepdim=True)
                    rollout = torch.matmul(attn, rollout)  # (1, 197, 197)

                rollout = rollout[:, 0] if data_from == "cls" else rollout[:, 1:].mean(dim=1) # (L, B, T)
                rollout = rollout / rollout.max(dim=-1, keepdim=True)[0]
                rollouts.append(rollout)
        rollouts = torch.stack(rollouts, dim=0)
        rollouts = rollouts[:, :, 1:]

        if reshape:
            rollouts = rollouts.reshape(-1, B, H, W) # L, B, H, W

    if ls_attentive:
        # attentive = saliencys[ls_attentive, :, 0] \
        #         if data_from == "cls" else saliencys[ls_attentive, :, 1:].mean(dim=2) # (L, B, T)
        attentive = saliencys[ls_attentive, :, 0] \
            if data_from == "cls" else saliencys[ls_attentive].mean(dim=2)  # (L, B, T)
        attentive = attentive[:, :, 1:]
        if reshape:
            attentive = attentive.reshape(-1, B, H, W)

    return rollouts, attentive



def data2PIL(datas):
    if type(datas) == torch.Tensor: # MAKE TO (..., H, W, 3)
        if datas.shape[-1] == 3:
            pils = datas
        elif len(datas.shape) == 2:
            pils = datas.unsqueeze(-1).detach().cpu()
        elif len(datas.shape) == 3:
            pils = datas.permute(1, 2, 0).detach().cpu()
        elif len(datas.shape) == 4:
            pils = datas.permute(0, 2, 3, 1).detach().cpu()
    elif type(datas) == np.ndarray:
        if len(datas.shape) == 2:
            datas = np.expand_dims(datas, axis=-1)
        # TODO NEED TO CHECK
        # if datas.max() <= 1:
        #     pils = cv2.cvtColor(datas, cv2.COLOR_BGR2RGB)  # 0.29ms
        # else:
        #     pils = datas

        # image = Image.fromarray(image)                 # 0.32ms
        pils = cv2.cvtColor(datas, cv2.COLOR_BGR2RGB)  # 0.29ms

    elif type(datas) == PIL.JpegImagePlugin.JpegImageFile \
         or type(datas) == PIL.Image.Image:
        pils = datas
    else:
        raise ValueError
    return pils


def drawImgPlot(datas:list, col=1, title:str=None, columns=None, showRows:list=None,
                output_dir='./', save_name=None, show=True, fmt=1,
                vis_x = False, vis_y = False, border=False, p=False):
    if type(datas) != list or 'shape' not in dir(datas[0]) : datas = [datas]

    if showRows is not None:
        for d in range(len(datas)):
            datas[d] = datas[d][showRows]
    data_num = len(datas[0]) * len(datas)
    row = (data_num - 1) // col + 1

    fig, axes = plt.subplots(nrows=row, ncols=col, squeeze=False)
    fig.set_size_inches(col * 8, row * 8)
    if title: fig.suptitle(title, fontsize=16)
    for i in range(data_num):
        r_num, c_num   = i // col, i % col
        dr_num, dc_num = i // len(datas), i % len(datas)
        data = data2PIL(datas[dc_num][dr_num])
        ax = axes[r_num][c_num]
        if "shape" not in dir(data): # IMAGE
            ax.imshow(data)
        elif data.shape[-1] == 3: #
            ax.imshow(data)
        elif data.shape[-1] == 1: #
            ax.imshow(data, cmap="gray")
            # sns.heatmap(data[:,:,0], annot=True, fmt=f".{fmt}f", cmap="Greys"
            #             , yticklabels=False, xticklabels=False, ax=ax, vmax=1.0, vmin=0.0)
        if not border:
            ax.set_axis_off()
        if columns:
            ax.set_title(columns[c_num] + str(r_num)) if len(columns) == col else ax.set_title(columns[i])
        if not vis_x: ax.xaxis.set_visible(False)
        if not vis_y: ax.yaxis.set_visible(False)
    plt.tight_layout()

    if output_dir and save_name:
        if not os.path.exists(output_dir):
            os.makedirs(output_dir, exist_ok=True)
        plt.savefig(os.path.join(output_dir, f'{save_name}.png'), transparent = True)
    if show:
        plt.show()

def overlay_mask(img: Image.Image, mask: Image.Image, colormap: str = "jet", alpha: float = 0.7) -> Image.Image:
    cmap = mpl.cm.get_cmap(colormap)
    # Resize mask and apply colormap
    overlay = mask.resize(img.size, resample=Image.BICUBIC)
    overlay = (255 * cmap(np.asarray(overlay) ** 2)[:, :, :3]).astype(np.uint8)
    # Overlay the image with the mask
    overlayed_img = Image.fromarray((alpha * np.asarray(img) + (1 - alpha) * overlay).astype(np.uint8))

    return overlayed_img



def overlay(imgs, attnsL, dataset="imagenet"):
    attnsL = to_leaf(to_tensor(attnsL))
    imgs   = to_leaf(to_tensor(imgs))
    if type(attnsL) == list:   attnsL = torch.stack(attnsL, 0)
    if len(attnsL.shape) == 2: attnsL = attnsL.unsqueeze(0) #
    if len(attnsL.shape) == 3: attnsL = attnsL.unsqueeze(0) # L, B, h, w
    if len(imgs.shape)   == 3: imgs = imgs.unsqueeze(0)     # B, C, H, W
    if dataset:
        imgs  = unNormalize(imgs, dataset)

    results = []
    for attns in attnsL:
        for img, attn in zip(imgs, attns):
            result = overlay_mask(to_pil_image(img), to_pil_image(normalization(attn, type=0), mode='F')) # (3, 224, 224), (1, 14, 14)
            # plt.imshow(overlay_mask(to_pil_image(dataset.unNormalize(samples)[0]), to_pil_image(normalization(a[:, 0]), mode='F'), alpha=0.5))
            results.append(result)
    return results

import copy
def unNormalize(image, dataset="imagenet"):
    # images : (B, C, H, W)
    if dataset == "imagenet":
        mean, std = [0.485, 0.456, 0.406], [0.229, 0.224, 0.225]
    elif dataset == "cifar":
        mean, std = [0.4914, 0.4822, 0.4465], [0.2023, 0.1994, 0.2010]
    else:
        return image

    result = copy.deepcopy(image)
    for c, (m, s) in enumerate(zip(mean, std)):
        result[:, c].mul_(s).add_(m)
    return result

def show_mask_on_image(img, mask):
    img = np.float32(img) / 255
    heatmap = cv2.applyColorMap(np.uint8(255 * mask), cv2.COLORMAP_JET)
    heatmap = np.float32(heatmap) / 255
    cam = heatmap + np.float32(img)
    cam = cam / np.max(cam)
    return np.uint8(255 * cam)